import dataclasses
from enum import Enum
from typing import Any
from typing import Dict
from typing import Optional
from typing import TypedDict


@dataclasses.dataclass(frozen=True)
class DatasetId:
    name: str
    path: str


@dataclasses.dataclass(frozen=True)
class DatasetMetadata:
    file_count: int


@dataclasses.dataclass(frozen=True)
class Dataset:
    dataset_id: DatasetId
    icat_dataset_id: int
    dataset_metadata: DatasetMetadata

    def as_dict(self) -> Dict[str, Any]:
        return dataclasses.asdict(self)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "Dataset":
        """Factory method to create a Dataset instance from a dictionary."""
        data = data.copy()
        data["dataset_id"] = DatasetId(**data["dataset_id"])
        data["dataset_metadata"] = DatasetMetadata(**data["dataset_metadata"])
        return cls(**data)


class ArchiveStatusType(Enum):
    ARCHIVING = "archiving"
    RESTORATION = "restoration"


class ArchiveStatusLevel(Enum):
    INFO = "info"
    WARNING = "warning"
    ERROR = "error"


class SampleMetadata(TypedDict):
    """
    Sample metadata information
    """

    id: str
    name: str
    linkedSampleAFormName: Optional[str]  # noqa: N815
    proposal: str
    beamline: str
    sessionStartDate: Optional[str]  # noqa: N815



import logging
import warnings
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

from icat_esrf_definitions.errors import ValidationWarning
from icat_esrf_definitions.models import IcatDatasetParameters
from pydantic import ValidationError

from .... import errors

logger = logging.getLogger(__name__)


def pydantic_serialization(
    parameters: Dict[str, Any], strict: bool = False
) -> Tuple[Dict[str, Any], List[str], Optional[Exception]]:
    """
    Pydantic serialization of dataset parameters.

    :raises IcatMetadataValidationError: If ``strict=True`` and the metadata is invalid.
    :emits IcatMetadataValidationWarning: If ``strict=False`` and the metadata is invalid.
    :emits IcatMetadataDeprecatedWarning: If metadata is deprecated.
    """
    parsed_parameters = parameters
    unknown_keys = []
    error = None
    captured_warnings = []

    try:
        try:
            try:
                parsed_parameters, unknown_keys, captured_warnings = (
                    _fully_pydantic_serialization(parameters)
                )
            except ValidationError as ex:
                logger.debug(
                    "Full Pydantic serialization failed; attempting partial serialization",
                    exc_info=ex,
                )
                try:
                    parsed_parameters, unknown_keys = _partial_pydantic_serialization(
                        parameters, ex
                    )
                    logger.debug("Partial Pydantic serialization succeeded")
                except Exception as inner_ex:
                    logger.debug(
                        "Partial Pydantic serialization failed; fallback to native string serialization",
                        exc_info=inner_ex,
                    )
                raise
        finally:
            for warning in captured_warnings:
                message = str(warning.message)
                if isinstance(warning, DeprecationWarning):
                    errors.warn(message, errors.IcatMetadataDeprecatedWarning)
                elif isinstance(warning, ValidationWarning):
                    errors.invalid_warn_or_raise(message, strict=strict)
                else:
                    errors.warn(message, errors.IcatMetadataValidationWarning)

    except Exception as ex:
        error = ex

    return parsed_parameters, unknown_keys, error


def _fully_pydantic_serialization(
    parameters: Dict[str, Any],
) -> Tuple[Dict[str, Any], List[str], List[Warning]]:
    """
    Pydantic serialization of all dataset parameters.

    Capture the following warnings

    - Pydantic `DeprecationWarning`
    - icat-esrf-definitions `ValidationWarning`
    """
    with warnings.catch_warnings(record=True) as capture_warnings:
        warnings.simplefilter("always", DeprecationWarning)
        warnings.simplefilter("always", ValidationWarning)
        model, unknown_keys = IcatDatasetParameters.from_icat_dict_known_keys(
            parameters
        )
        parsed_parameters = model.to_icat_dict()
    return parsed_parameters, unknown_keys, capture_warnings


def _partial_pydantic_serialization(
    parameters: Dict[str, Any], error: ValidationError
) -> Tuple[Dict[str, Any], List[str]]:
    """
    Pydantic serialization of the dataset parameter that did not cause a validation error.

    Hide Pydantic `DeprecationWarning`.
    """
    invalid_names = IcatDatasetParameters.icat_field_names_with_validation_error(error)
    valid_parameters = {k: v for k, v in parameters.items() if k not in invalid_names}
    invalid_parameters = {k: v for k, v in parameters.items() if k in invalid_names}

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        model, unknown_keys = IcatDatasetParameters.from_icat_dict_known_keys(
            valid_parameters
        )
        parsed_valid_parameters = model.to_icat_dict()
        parsed_parameters = {**parsed_valid_parameters, **invalid_parameters}

    return parsed_parameters, unknown_keys



import base64
import random
from typing import List
from typing import Optional

import requests
from stompest.config import StompConfig
from stompest.error import StompConnectionError
from stompest.protocol import StompSession
from stompest.protocol import StompSpec
from stompest.sync import Stomp

from ..utils.url import normalize_url


class IcatMessagingClient:
    """Client for the ICAT message broker.

    The message broker is currently ActiveMQ, a message
    broker using the STOMP protocol. It also has a REST
    server for monitoring the broker status.
    """

    DEFAULT_SCHEME = "tcp"
    DEFAULT_PORT = 61613
    MONITOR_SCHEME = "http"
    DEFAULT_MONITOR_PORT = 8778
    MONITOR_USER = "user"
    MONITOR_PWD = "user"

    def __init__(
        self,
        queue_urls: List[str],
        queue_name: str,
        monitor_port: Optional[int] = None,
        timeout: Optional[float] = None,
    ):
        urls = [
            normalize_url(
                url, default_scheme=self.DEFAULT_SCHEME, default_port=self.DEFAULT_PORT
            )
            for url in queue_urls
        ]
        self.__queue_urls = urls
        self.__max_version = StompSpec.VERSION_1_1
        self.__client = self._create_client()
        self._socket_timeout = None
        if timeout is None:
            timeout = 5
        self._connect_timeout = timeout

        self.__send_destination = "/queue/" + queue_name
        self.__send_headers = {
            "persistent": "true",
            StompSpec.ACK_HEADER: StompSpec.ACK_CLIENT_INDIVIDUAL,
        }

        if not monitor_port:
            monitor_port = self.DEFAULT_MONITOR_PORT
        self.__consumer_count_url = f"{self.MONITOR_SCHEME}://{{host}}:{monitor_port}/api/jolokia/read/org.apache.activemq:type=Broker,brokerName=metadata,destinationType=Queue,destinationName={queue_name}/ConsumerCount"
        self.__jolokia_headers = {
            "Authorization": b"Basic "
            + base64.b64encode(f"{self.MONITOR_USER}:{self.MONITOR_PWD}".encode())
        }

    def _create_client(self):
        random.shuffle(self.__queue_urls)
        failover = ",".join(self.__queue_urls)
        url = f"failover:({failover})?maxReconnectAttempts=3,initialReconnectDelay=250,maxReconnectDelay=1000"
        return Stomp(StompConfig(url, version=self.__max_version))

    def reconnect(self):
        self.disconnect()
        self.__client = self._create_client()
        self._connect()

    def disconnect(self):
        try:
            self.__client.disconnect()
        except StompConnectionError:
            self.__client.close(flush=True)

    def _connect(self):
        self.__client.connect(
            versions=[self.__max_version],
            connectTimeout=self._socket_timeout,
            connectedTimeout=self._connect_timeout,
        )

    @property
    def _connected_client(self):
        if self.__client.session.state != StompSession.CONNECTED:
            self._connect()
        return self.__client

    @property
    def _host(self):
        return self._connected_client._transport.host

    def send(self, data: bytes):
        try:
            self._send(data)
        except StompConnectionError:
            self.reconnect()
            self._send(data)

    def _send(self, data: bytes):
        self._connected_client.send(
            self.__send_destination, body=data, headers=self.__send_headers
        )

    @property
    def _consumer_count(self):
        url = self.__consumer_count_url.format(host=self._host)
        response = requests.get(url, headers=self.__jolokia_headers)
        if not response.ok:
            raise RuntimeError(
                response, "Failed to retrieve the ActiveMQ consumer count"
            )
        response = response.json()
        return response["value"]

    def check_health(self):
        """Raises an exception when:
        - not connected
        - no message consumers
        """
        state = self._connected_client.session.state
        if state != StompSession.CONNECTED:
            raise RuntimeError(
                "The connection with the message broker is " + str(state).upper()
            )
        if self._consumer_count < 1:
            raise RuntimeError("The message broker has no consumers")



import warnings

import pytest
from esrf_ontologies import technique

from ..errors import IcatMetadataFutureWarning
from ..metadata.definitions import load_icat_fields
from .fixtures.icat import *  # noqa F401

pytest.register_assert_rewrite("pyicat_plus.tests.utils.compare")


@pytest.fixture
def icat_namespace():
    metadict = dict()

    def getter(key):
        return metadict[key]

    def setter(key, value):
        metadict[key] = value

    icat_fields = load_icat_fields()
    metadata = icat_fields.namespace(getter=getter, setter=setter)

    return icat_fields, metadata, metadict


@pytest.fixture
def ignore_icat_message_warnings():
    """Ignore expected ICAT serialization warnings."""
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=IcatMetadataFutureWarning)
        yield


@pytest.fixture(scope="session")
def esrfet_version() -> str:
    return technique.get_ontology_version_number()



import socket
from collections.abc import Mapping
from importlib.metadata import version as get_version

import numpy

release = get_version("pyicat_plus")


def assert_equal_hdf5_dict(adict: Mapping, expected: Mapping):
    """Check that HDF5 content is the same."""
    stack = [(adict, expected)]
    while stack:
        adict, expected = stack.pop(0)
        existing_keys = set(adict.keys())
        expected_keys = set(expected.keys())
        missing_keys = expected_keys - existing_keys
        unexpected_keys = existing_keys - expected_keys
        if missing_keys:
            raise AssertionError(f"missing keys: {missing_keys}")
        if unexpected_keys:
            raise AssertionError(f"unexpected keys: {unexpected_keys}")
        for k, v in expected.items():
            if isinstance(v, Mapping):
                stack.append((adict[k], v))
            elif isinstance(v, numpy.ndarray) and v.size > 1:
                assert adict[k].shape == v.shape
                assert adict[k].dtype == v.dtype
                if adict[k].dtype != object:
                    assert all(
                        numpy.isnan(adict[k].flatten()) == numpy.isnan(v.flatten())
                    )
                    mask = numpy.logical_not(numpy.isnan(v.flatten()))
                    assert all((adict[k].flatten() == v.flatten())[mask])
                else:
                    assert all(adict[k].flatten() == v.flatten())
            elif isinstance(v, (list, tuple)):
                assert list(adict[k]) == list(v)
            else:
                assert adict[k] == v


def assert_equal_dataset_message(actual: dict, expected: dict) -> None:
    """Check that ICAT ingester dataset messages are the same."""

    # Cannot know the exact time
    _set_expected_to_actual_value(actual, expected, "startDate")
    _set_expected_to_actual_value(actual, expected, "endDate")
    _set_expected_to_actual_value(actual, expected, "parameter", "startDate")
    _set_expected_to_actual_value(actual, expected, "parameter", "endDate")

    # Always set when missing
    _set_expected_value(expected, "pyicat-plus_v" + release, "parameter", "software")
    _set_expected_value(expected, socket.getfqdn(), "parameter", "machine")

    # XSD defaults (excluding None)
    _set_expected_value(expected, [], "datafile")
    _set_expected_value(expected, [], "sample", "parameter")

    # Parameter list can be in any order
    _sort_parameters(actual)
    _sort_parameters(expected)

    assert actual == expected


def assert_equal_investigation_message(actual: dict, expected: dict) -> None:
    """Check that ICAT ingester investigation messages are the same."""
    _set_expected_to_actual_value(actual, expected, "startDate")

    assert actual == expected


def _set_expected_value(expected: dict, value: str, *keys) -> None:
    """Set expected value when not provided."""
    expected_parent = expected

    for i, key in enumerate(keys[:-1]):
        if key == "parameter":
            next_key = keys[i + 1]

            param_list = expected_parent.setdefault("parameter", [])

            if not isinstance(param_list, list):
                raise TypeError(
                    f"Expected list at '{'.'.join(keys[:i+1])}', got {type(param_list).__name__}"
                )

            # Do not overwrite existing parameter
            for param in param_list:
                if param.get("name") == next_key:
                    return

            param_list.append({"name": next_key, "value": value})
            return

        expected_parent = expected_parent.setdefault(key, {})

    _ = expected_parent.setdefault(keys[-1], value)


def _set_expected_to_actual_value(actual: dict, expected: dict, *keys: str) -> None:
    """Set expected value to actual value when not provided."""
    if not keys:
        return

    value = actual
    for i, key in enumerate(keys):
        if not isinstance(value, dict):
            raise TypeError(
                f"Expected dict while traversing path '{'.'.join(keys)}', "
                f"but got {type(value).__name__} at key '{key}'."
            )

        if key not in value:
            path = ".".join(keys[:i]) or "<root>"
            raise KeyError(
                f"Missing key '{key}' in '{list(value)}' under '{path}' "
                f"while traversing '{'.'.join(keys)}'"
            )

        value = value[key]

        if key == "parameter":
            # parameter is a list of {"name": ..., "value": ...}
            next_key = keys[i + 1]

            if not isinstance(value, list):
                raise TypeError(
                    f"Expected list at '{'.'.join(keys[:i+1])}', "
                    f"got {type(value).__name__}"
                )

            match = next(
                (p["value"] for p in value if p.get("name") == next_key),
                None,
            )

            if match is None:
                raise KeyError(
                    f"Parameter '{next_key}' not found in "
                    f"{[p.get('name') for p in value]}"
                )

            value = match
            break

    _set_expected_value(expected, value, *keys)


def _sort_parameters(d: dict) -> None:
    """Recursively sort all 'parameter' lists by their 'name' field in a nested dict."""
    if not isinstance(d, Mapping):
        return

    for key, value in d.items():
        if key == "parameter" and isinstance(value, list):
            # Sort the list of dicts by 'name'
            value.sort(key=lambda p: p.get("name", ""))
        elif isinstance(value, dict):
            _sort_parameters(value)
        elif isinstance(value, list):
            for item in value:
                if isinstance(item, dict):
                    _sort_parameters(item)



import datetime
from typing import Any
from typing import Dict
from typing import Optional
from typing import Union

from xsdata.formats.dataclass.parsers import DictDecoder

from ... import errors
from ..models.xsd.investigation import Investigation
from . import xml


def serialize_investigation_message(
    beamline: str,
    proposal: str,
    start_datetime: Optional[Union[datetime.datetime, str]] = None,
    end_datetime: Optional[Union[datetime.datetime, str]] = None,
    investigation_id: Optional[str] = None,
    strict: bool = False,
    indent: Optional[str] = None,
) -> bytes:
    """
    :raises IcatMetadataValidationError: If ``strict=True`` and the metadata is invalid.
    :emits IcatMetadataValidationWarning: If ``strict=False`` and the metadata is invalid.
    :emits IcatMetadataDeprecatedWarning: If metadata is deprecated.
    """
    if start_datetime is None:
        start_datetime = datetime.datetime.now().astimezone()

    if end_datetime is not None:
        # This was added for the unit tests but fails XSD validation.
        # Not used in production.
        message = "ICAT Investigation schema does not have 'end_datetime'. Ignore it."
        errors.invalid_warn_or_raise(message, strict=strict)

    # XML message root
    root = {
        "experiment": proposal,
        "instrument": beamline,
        "startDate": start_datetime,
        "investigationId": investigation_id,  # Trick XSD validation when None
    }

    # Serialize before XSD serialization
    root = xml.serialize_root(root)

    # Serialize with XSD validation
    return raw_serialize_investigation_message(root, indent=indent)


def raw_serialize_investigation_message(
    data: Dict[str, Any], indent: Optional[str] = None
) -> bytes:
    """Serialize message with XSD 'investigation' validation."""
    model = DictDecoder().decode(data, Investigation)
    return xml.serialize_message(model, indent=indent)



from typing import Any

from ..metadata.namespace_wrapper import NamespaceWrapper


def _wrap_dict(adict: dict) -> NamespaceWrapper:
    def getter(key: str) -> Any:
        item = adict.get(key, None)
        if item is None:
            raise AttributeError(key)
        if isinstance(item, dict):
            return _wrap_dict(item)
        return item

    def setter(key: str, value: Any) -> Any:
        adict[key] = value

    return NamespaceWrapper(
        property_names=list(adict),
        getter=getter,
        setter=setter,
    )


def test_namespace_str():
    adict = {"field1": 1, "group1": {"field2": 2, "group2": {}}}
    namespace = _wrap_dict(adict)
    assert str(namespace) == "NameSpace:\n field1\n group1"
    assert str(namespace.group1) == "NameSpace:\n field2\n group2"
    assert str(namespace.group1.group2) == "NameSpace: <empty>"



import json
from pathlib import Path

from PIL import Image

from ..metadata.drac_report_builder import DracReportBuilder


def test_escape_characters():
    report = DracReportBuilder(dataset_type="demo")
    escaped = report.escape_characters("AÅ\nB° <&>")
    assert "&Aring;" in escaped
    assert "&deg;" in escaped
    assert "<br>" in escaped
    assert "&lt;&amp;&gt;" in escaped


def test_render_json_and_html(tmp_path):
    report = DracReportBuilder(dataset_type="demo", dataset_title="Demo Report")
    report.add_info("Info line")
    report.add_warning("Warn line")
    report.add_table(
        title="Sample Table",
        columns=["Name", "Value"],
        data=[["Alpha", 1], ["Beta", 2]],
    )
    log_path = tmp_path / "log.txt"
    log_path.write_text("Log <unsafe>", encoding="utf-8")
    report.add_log_file("Log", "View log", str(log_path))

    json_path = Path(report.render_json(str(tmp_path)))
    assert json_path.exists()
    data = json.loads(json_path.read_text(encoding="utf-8"))
    assert data["title"] == "Demo Report"
    assert [item["type"] for item in data["items"]] == [
        "info",
        "warning",
        "table",
        "logFile",
    ]

    html_path = Path(report.render_html(str(tmp_path)))
    assert html_path.exists()
    html_text = html_path.read_text(encoding="utf-8")
    assert "Demo Report" in html_text
    assert "Info line" in html_text
    assert "Warn line" in html_text
    assert "View log" in html_text
    assert "<table" in html_text

    log_htmls = sorted(tmp_path.glob("Log*.html"))
    assert log_htmls
    log_html_text = log_htmls[0].read_text(encoding="utf-8")
    assert "&lt;unsafe&gt;" in log_html_text


def test_render_image_in_html(tmp_path):
    image_path = tmp_path / "sample.png"
    Image.new("RGB", (8, 6), color=(255, 0, 0)).save(image_path)

    report = DracReportBuilder(dataset_type="demo", dataset_title="Demo Report")
    report.add_image(str(image_path), image_title="Sample Image")

    json_path = Path(report.render_json(str(tmp_path)))
    data = json.loads(json_path.read_text(encoding="utf-8"))
    assert data["items"][-1]["type"] == "image"
    assert data["items"][-1]["value"]

    html_path = Path(report.render_html(str(tmp_path)))
    assert html_path.exists()
    html_text = html_path.read_text(encoding="utf-8")
    assert "<img" in html_text
    assert "Sample_Image.png" in html_text

    rendered_image = tmp_path / "Sample_Image.png"
    assert rendered_image.exists()
    assert rendered_image.stat().st_size > 0


def test_render_image_with_thumbnail(tmp_path):
    image_path = tmp_path / "sample.png"
    thumbnail_path = tmp_path / "thumbnail.png"
    Image.new("RGB", (8, 6), color=(255, 0, 0)).save(image_path)
    Image.new("RGB", (4, 3), color=(0, 255, 0)).save(thumbnail_path)

    report = DracReportBuilder(dataset_type="demo", dataset_title="Demo Report")
    report.add_image(
        str(image_path),
        image_title="Sample Image",
        path_to_thumbnail_image=str(thumbnail_path),
    )

    json_path = Path(report.render_json(str(tmp_path)))
    data = json.loads(json_path.read_text(encoding="utf-8"))
    assert data["items"][-1]["type"] == "image"
    assert data["items"][-1]["thumbnailValue"]

    html_path = Path(report.render_html(str(tmp_path)))
    assert html_path.exists()
    html_text = html_path.read_text(encoding="utf-8")
    assert "<img" in html_text
    assert "Sample_Image_thumbnail.png" in html_text

    rendered_thumbnail = tmp_path / "Sample_Image_thumbnail.png"
    assert rendered_thumbnail.exists()
    assert rendered_thumbnail.stat().st_size > 0



from datetime import datetime
from datetime import timedelta
from pathlib import Path

from ..apps import store_from_file
from ..apps import store_processed
from ..apps import store_raw
from .utils import compare


def test_store_raw(icat_metadata_client, icat_backend):
    _, messages = icat_metadata_client

    argv = [
        "",
        "--beamline",
        "id00",
        "--proposal",
        "hg123",
        "--dataset",
        "datasetname",
        "--path",
        _dummy_path("dataset"),
        "--sample",
        "samplename",
        "-p",
        "InstrumentBeam_distance=0.01",
        "-p",
        "InstrumentBeam_incident_energy=[10, 10.1, 10.2]",
        "--queue",
        icat_backend["metadata_urls"][0],
    ]
    store_raw.main(argv)

    message = messages.get(timeout=10)
    assert messages.empty()

    expected = {
        "complete": True,
        "instrument": "id00",
        "investigation": "hg123",
        "location": _dummy_path("dataset"),
        "name": "datasetname",
        "parameter": [
            {"name": "complete", "value": "true"},
            {"name": "proposal", "value": "hg123"},
            {"name": "location", "value": str(_dummy_path("dataset"))},
            {"name": "Sample_name", "value": "samplename"},
            {"name": "datasetName", "value": "datasetname"},
            {"name": "InstrumentBeam_distance", "value": "0.01"},
            {"name": "InstrumentBeam_incident_energy", "value": "10.0 10.1 10.2"},
        ],
        "sample": {"name": "samplename"},
    }
    compare.assert_equal_dataset_message(message, expected)


def test_store_raw_metadatafile(icat_metadata_client, icat_backend, tmpdir):
    metadatafile = str(tmpdir / "metadata.txt")

    with open(metadatafile, "w") as f:
        f.write("InstrumentBeam_distance=0.01\n")
        f.write("InstrumentBeam_incident_energy=[10, 10.1, 10.2]\n")

    _, messages = icat_metadata_client

    argv = [
        "",
        "--beamline",
        "id00",
        "--proposal",
        "hg123",
        "--dataset",
        "datasetname",
        "--path",
        _dummy_path("dataset"),
        "--sample",
        "samplename",
        "--metadatafile",
        metadatafile,
        "--queue",
        icat_backend["metadata_urls"][0],
    ]
    store_raw.main(argv)

    message = messages.get(timeout=10)
    assert messages.empty()

    expected = {
        "complete": True,
        "instrument": "id00",
        "investigation": "hg123",
        "location": _dummy_path("dataset"),
        "name": "datasetname",
        "parameter": [
            {"name": "complete", "value": "true"},
            {"name": "proposal", "value": "hg123"},
            {"name": "location", "value": str(_dummy_path("dataset"))},
            {"name": "Sample_name", "value": "samplename"},
            {"name": "datasetName", "value": "datasetname"},
            {"name": "InstrumentBeam_distance", "value": "0.01"},
            {"name": "InstrumentBeam_incident_energy", "value": "10.0 10.1 10.2"},
        ],
        "sample": {"name": "samplename"},
    }
    compare.assert_equal_dataset_message(message, expected)


def test_store_processed(icat_metadata_client, icat_backend):
    _, messages = icat_metadata_client

    argv = [
        "",
        "--beamline",
        "id00",
        "--proposal",
        "hg123",
        "--dataset",
        "datasetname",
        "--path",
        _dummy_path("processed"),
        "--sample",
        "samplename",
        "-p",
        "InstrumentBeam_distance=0.01",
        "-p",
        "InstrumentBeam_incident_energy=[10, 10.1, 10.2]",
        "--queue",
        icat_backend["metadata_urls"][0],
        "--raw",
        _dummy_path("dataset1"),
        "--raw",
        _dummy_path("dataset2"),
    ]
    store_processed.main(argv)

    expected = {
        "complete": True,
        "instrument": "id00",
        "investigation": "hg123",
        "location": _dummy_path("processed"),
        "name": "datasetname",
        "parameter": [
            {"name": "complete", "value": "true"},
            {"name": "proposal", "value": "hg123"},
            {"name": "location", "value": str(_dummy_path("processed"))},
            {"name": "Sample_name", "value": "samplename"},
            {"name": "datasetName", "value": "datasetname"},
            {"name": "InstrumentBeam_distance", "value": "0.01"},
            {"name": "InstrumentBeam_incident_energy", "value": "10.0 10.1 10.2"},
            {
                "name": "input_datasets",
                "value": f"{_dummy_path('dataset1')},{_dummy_path('dataset2')}",
            },
        ],
        "sample": {"name": "samplename"},
    }
    message = messages.get(timeout=10)
    assert messages.empty()

    compare.assert_equal_dataset_message(message, expected)


def test_store_from_file(icat_main_client, icat_backend, tmp_path):
    client, messages = icat_main_client
    icat_dir = tmp_path / "__icat__"
    icat_dir.mkdir()

    start_date = datetime.now()
    client.start_investigation(
        beamline="id00",
        proposal="hg123",
        start_datetime=start_date,
    )
    messages.get(timeout=10)

    dataset_start = start_date + timedelta(hours=1)
    dataset_end = start_date + timedelta(hours=2)
    file1 = icat_dir / "dataset1.xml"
    file1.write_text(
        _build_dataset_xml(
            "0001", str(icat_dir / "sample_0001"), dataset_start, dataset_end
        )
    )
    argv = [
        "store_from_file_script",
        str(file1),
        "--queue",
        icat_backend["metadata_urls"][0],
    ]
    store_from_file.main(argv)

    message = messages.get(timeout=10)
    expected = {
        "complete": True,
        "name": "0001",
        "instrument": "id00",
        "investigation": "hg123",
        "location": str(icat_dir / "sample_0001"),
        "parameter": [
            {"name": "startDate", "value": dataset_start.isoformat()},
            {"name": "endDate", "value": dataset_end.isoformat()},
            {"name": "Sample_name", "value": "sample"},
            {"name": "machine", "value": "machine"},
            {"name": "software", "value": "test"},
        ],
        "sample": {"name": "sample", "parameter": [{}]},
    }
    compare.assert_equal_dataset_message(message, expected)
    assert messages.empty()


def _build_dataset_xml(name, location, start_date, end_date):
    return (
        f'<tns:dataset xmlns:tns="http://www.esrf.fr/icat" complete="true">'
        f"<tns:investigation>hg123</tns:investigation>"
        f"<tns:instrument>id00</tns:instrument>"
        f"<tns:name>{name}</tns:name>"
        f"<tns:location>{location}</tns:location>"
        f"<tns:parameter><tns:name>Sample_name</tns:name><tns:value>sample</tns:value></tns:parameter>"
        f"<tns:parameter><tns:name>startDate</tns:name><tns:value>{start_date.isoformat()}</tns:value></tns:parameter>"
        f"<tns:parameter><tns:name>endDate</tns:name><tns:value>{end_date.isoformat()}</tns:value></tns:parameter>"
        f"<tns:parameter><tns:name>machine</tns:name><tns:value>machine</tns:value></tns:parameter>"
        f"<tns:parameter><tns:name>software</tns:name><tns:value>test</tns:value></tns:parameter>"
        f"<tns:startDate>{start_date.isoformat()}</tns:startDate>"
        f"<tns:endDate>{end_date.isoformat()}</tns:endDate>"
        f"<tns:sample><tns:name>sample</tns:name><tns:parameter></tns:parameter></tns:sample>"
        f"</tns:dataset>"
    )


def _dummy_path(dirname: str) -> str:
    return str(Path.home() / dirname)



import json
import logging
import os
from typing import Optional
from uuid import uuid4

logger = logging.getLogger("ICATPLUS SERVER")


class IcatSession:
    def __init__(self, root_dir: Optional[str] = None):
        if root_dir is None:
            root_dir = "."
        if root_dir:
            os.makedirs(root_dir, exist_ok=True)
        self._root_dir = root_dir
        self._session_file = os.path.join(root_dir, "session.json")

    def login(self, credentials: dict) -> dict:
        if credentials["password"] != "correct":
            raise PermissionError("Password incorrect")
        session_id = str(uuid4())
        session_data = {"sessionId": session_id}
        with open(self._session_file, "w") as f:
            json.dump(session_data, f)
        logger.info("Login session id = %s", session_id)
        return session_data

    @property
    def session_data(self) -> Optional[dict]:
        if os.path.exists(self._session_file):
            with open(self._session_file, "r") as f:
                return json.load(f)

    @property
    def session_id(self) -> Optional[str]:
        data = self.session_data
        if data:
            session_id = data["sessionId"]
        else:
            session_id = None
        logger.info("Current session id = %s", session_id)
        return session_id

    def is_allowed(self, session_id: str) -> bool:
        logger.info("Validate session id = %s", session_id)
        return session_id == self.session_id



"""
Comparison of ISO 8601 timezone formats using Zulu (Z) and numeric offset (+hh:mm) notation.

UTC timezone:

- XmlDateTime.__str__()                 -> 2023-10-19T21:53:07.834537Z
- BaseModel.model_dump(mode="json")     -> 2023-10-19T21:53:07.834537Z
- datetime.datetime.isoformat()         -> 2023-10-19T21:53:07.834537+00:00

Non-UTC timezone:

- XmlDateTime.__str__()                 -> 2023-10-19T21:53:07.834537+02:00
- BaseModel.model_dump(mode="json")     -> 2023-10-19T21:53:07.834537+02:00
- datetime.datetime.isoformat()         -> 2023-10-19T21:53:07.834537+02:00
"""

import datetime
from typing import Union

from xsdata.models.datatype import XmlDateTime


def with_timezone_offset_format(dt: Union[XmlDateTime, datetime.datetime, str]) -> str:
    if isinstance(dt, str):
        dt = XmlDateTime.from_string(dt)
    if isinstance(dt, XmlDateTime):
        return dt.to_datetime().isoformat()
    if isinstance(dt, datetime.datetime):
        return dt.isoformat()
    raise TypeError(
        f"Expected XmlDateTime, datetime.datetime or str, got {type(dt).__name__}"
    )


def with_timezone_zulu_format(dt: Union[XmlDateTime, datetime.datetime, str]) -> str:
    if isinstance(dt, str):
        dt = XmlDateTime.from_string(dt)
    if isinstance(dt, XmlDateTime):
        return str(dt)
    if isinstance(dt, datetime.datetime):
        return str(XmlDateTime.from_datetime(dt))
    raise TypeError(
        f"Expected XmlDateTime, datetime.datetime or str, got {type(dt).__name__}"
    )


def deserialize_datetime(dt: str) -> datetime.datetime:
    # In CI at UTC python<=3.10 fromisoformat raises an exception for Zulu timezone format.
    # Since python 3.11 most ISO 8601 formats can be parsed.
    # https://docs.python.org/3/whatsnew/3.11.html#datetime
    return XmlDateTime.from_string(dt).to_datetime()



import os

import h5py
import pytest

from ..utils import path_utils
from ..utils import raw_data


def test_get_dataset_filters_should_raise_exception_if_invalid_format(tmp_path):
    with pytest.raises(
        NotImplementedError, match="Raw data format 'unknown' is not supported"
    ):
        raw_data.get_dataset_filters(
            raw_root_dir=str(tmp_path), raw_data_format="unknown"
        )


@pytest.mark.parametrize("format", ["esrfv1", "esrfv2", "esrfv3", "id16bspec"])
def test_get_dataset_filters_should_return_expected_path_for_esrf_id16bspec_format(
    tmp_path,
    format,
):
    result = raw_data.get_dataset_filters(
        raw_root_dir=str(tmp_path), raw_data_format=format
    )
    expected_path = os.path.join(str(tmp_path), "*", "*")
    assert result == [path_utils.markdir(expected_path)]


def test_get_dataset_filters_should_return_expected_path_for_mx_format(tmp_path):
    create_folder(tmp_path / "collection")
    create_folder(tmp_path / "workflow", False, False)
    create_folder(tmp_path / "workflow" / "line")
    create_folder(tmp_path / "workflow" / "mesh", False, True)

    result = raw_data.get_dataset_filters(
        raw_root_dir=str(tmp_path), raw_data_format="mx"
    )
    expected_paths = {
        os.path.join(str(tmp_path), "collection", ""),
        os.path.join(str(tmp_path), "workflow", "line", ""),
    }

    assert set(result) == expected_paths


def create_folder(folder_path, with_metadata=True, with_master=True):
    folder_path.mkdir(parents=True, exist_ok=True)
    if with_metadata:
        metadata_path = folder_path / "metadata.json"
        with open(metadata_path, "w"):
            pass
    if with_master:
        master_h5_path = folder_path / "master.h5"
        with h5py.File(master_h5_path, "w"):
            pass



import json

import pytest


def send_update_metadata(client, value):
    client.send_update_metadata(
        proposal=value["proposal"],
        beamline=value["beamline"],
        dataset_paths=value["datasetPaths"],
        metadata_value=value["metadataValue"],
        metadata_name=value["metadataName"],
    )


def test_send_update_metadata_sample(icat_update_metadata_client):
    client, messages = icat_update_metadata_client
    client.check_health()
    expected = {
        "proposal": "sc415",
        "beamline": "id00",
        "datasetPaths": ["/data/visitor/sc415/id00/20230101/RAW_DATA/sample/dataset"],
        "metadataName": "Sample_name",
        "metadataValue": "newsamplename",
    }
    send_update_metadata(client, expected)
    message = messages.get(timeout=10)
    message = json.loads(message)
    assert message == expected
    assert messages.empty()


def test_send_update_metadata_sample_with_multiple_datasets(
    icat_update_metadata_client,
):
    client, messages = icat_update_metadata_client
    client.check_health()
    expected = {
        "proposal": "sc415",
        "beamline": "id00",
        "datasetPaths": [
            "/data/visitor/sc415/id00/20230101/RAW_DATA/sample/dataset1",
            "/data/visitor/sc415/id00/20230101/RAW_DATA/sample/dataset2",
        ],
        "metadataName": "Sample_name",
        "metadataValue": "newsamplename",
    }
    send_update_metadata(client, expected)
    message = messages.get(timeout=10)
    message = json.loads(message)
    assert message == expected
    assert messages.empty()


def test_send_update_metadata_raises_error_if_proposal_is_missing(
    icat_update_metadata_client,
):
    client, messages = icat_update_metadata_client
    with pytest.raises(AssertionError, match="ICAT requires the proposal"):
        client.send_update_metadata(
            proposal=None,
            beamline="id00",
            dataset_paths="/data/visitor/sc415/id00/20230101/RAW_DATA/sample/dataset/",
            metadata_value="newsamplename",
            metadata_name="Sample_name",
        )


def test_send_update_metadata_raises_error_if_beamline_is_missing(
    icat_update_metadata_client,
):
    client, messages = icat_update_metadata_client
    with pytest.raises(AssertionError, match="ICAT requires the beamline"):
        client.send_update_metadata(
            proposal="sc415",
            beamline=None,
            dataset_paths="/data/visitor/sc415/id00/20230101/RAW_DATA/sample/dataset/",
            metadata_value="newsamplename",
            metadata_name="Sample_name",
        )


def test_send_update_metadata_raises_error_if_dataset_paths_is_missing(
    icat_update_metadata_client,
):
    client, messages = icat_update_metadata_client
    with pytest.raises(AssertionError, match="ICAT requires the dataset path"):
        client.send_update_metadata(
            proposal="sc415",
            beamline="id00",
            dataset_paths=None,
            metadata_value="newsamplename",
            metadata_name="Sample_name",
        )


def test_send_update_metadata_raises_error_if_dataset_paths_is_empty(
    icat_update_metadata_client,
):
    client, messages = icat_update_metadata_client
    with pytest.raises(AssertionError, match="ICAT requires the dataset path"):
        client.send_update_metadata(
            proposal="sc415",
            beamline="id00",
            dataset_paths=[],
            metadata_value="newsamplename",
            metadata_name="Sample_name",
        )


def test_send_update_metadata_raises_error_if_metadata_value_is_missing(
    icat_update_metadata_client,
):
    client, messages = icat_update_metadata_client
    with pytest.raises(AssertionError, match="ICAT requires the metadata value"):
        client.send_update_metadata(
            proposal="sc415",
            beamline="id00",
            dataset_paths="/data/visitor/sc415/id00/20230101/RAW_DATA/sample/dataset/",
            metadata_value=None,
            metadata_name="Sample_name",
        )


def test_send_update_metadata_raises_error_if_metadata_name_is_missing(
    icat_update_metadata_client,
):
    client, messages = icat_update_metadata_client
    with pytest.raises(AssertionError, match="ICAT requires the metadata name"):
        client.send_update_metadata(
            proposal="sc415",
            beamline="id00",
            dataset_paths="/data/visitor/sc415/id00/20230101/RAW_DATA/sample/dataset/",
            metadata_value="newsamplename",
            metadata_name=None,
        )



import datetime
import logging
import os
from typing import List
from typing import Optional
from typing import Sequence
from typing import Union
from urllib.parse import urljoin

import numpy
import requests

from .._icat_messaging.utils import datetime_utils
from ..concurrency import QueryPool
from ..utils.maxsizedict import MaxSizeDict
from ..utils.url import normalize_url
from . import defaults
from .types import Dataset
from .types import DatasetId
from .types import DatasetMetadata

logger = logging.getLogger(__name__)

_DEFAULT_START_TIME = datetime.time(hour=8)


class IcatInvestigationClient:
    """Client for the investigation part of the ICAT+ REST API.

    An "investigation" is a time slot assigned to a particular proposal
    at a particular beamline.

    REST API docs:
    https://icatplus.esrf.fr/api-docs/

    The ICAT+ server project:
    https://gitlab.esrf.fr/icat/icat-plus/-/blob/master/README.md
    """

    DEFAULT_SCHEME = "https"

    def __init__(
        self,
        url: str,
        api_key: Optional[str] = None,
        timeout: Optional[float] = None,
    ):
        if api_key is None:
            api_key = defaults.ELOGBOOK_TOKEN
        url = normalize_url(url, default_scheme=self.DEFAULT_SCHEME)

        path = f"dataacquisition/{api_key}/investigation"
        query = "?instrumentName={beamline}&investigationName={proposal}"
        self._investigation_url = urljoin(url, path + query)

        path = f"dataacquisition/{api_key}/dataset"
        query = "?investigationId={investigation_id}"
        self._dataset_url = urljoin(url, path + query)

        self.raise_error = False
        self.__query_pool = QueryPool(timeout=timeout, maxqueries=20)
        self.__investigation_info = MaxSizeDict(maxsize=20)

    @property
    def timeout(self):
        return self.__query_pool.timeout

    @timeout.setter
    def timeout(self, value: Optional[float] = None):
        self.__query_pool.timeout = value

    def _get_with_response_parsing(
        self, url: str, timeout: Optional[float] = None
    ) -> Optional[list]:
        """Return `None` means the information is not available at this moment.
        An empty list means that an error has occured on the server side or an
        actual empty list is returned.
        """
        try:
            response = self.__query_pool.execute(
                requests.get, args=(url,), timeout=timeout, default=None
            )
        except requests.exceptions.ReadTimeout:
            return None
        except Exception as e:
            if self.raise_error:
                raise
            logger.exception(e)
            return None
        if response is None:
            return None
        if self.raise_error:
            response.raise_for_status()
        elif not response.ok:
            logger.error("%s: %s", response, response.text)
        if response.ok:
            return response.json()
        else:
            return list()

    def investigation_info(
        self,
        beamline: str,
        proposal: str,
        date: Optional[Union[datetime.datetime, datetime.date]] = None,
        allow_open_ended: bool = True,
        timeout: Optional[float] = None,
    ) -> Optional[dict]:
        """An investigation is defined by a time slot. Find an investigation (if any)
        for a beamline, proposal and date ("now" when not provided). When there is
        more than one investigation, select the closest one started before or at the date.
        If there is no such investigation, get the closest investigation which starts after the date.
        """
        investigation_key = beamline, proposal, date
        ninfo = self.__investigation_info.get(investigation_key)
        if ninfo is not None:
            return ninfo

        # Get all investigations for this proposal and beamline
        url = self._investigation_url.format(beamline=beamline, proposal=proposal)
        investigations = self._get_with_response_parsing(url, timeout=timeout)
        if investigations is None:
            return None  # not available at the moment

        # Select investigation
        investigation = _select_investigation(
            investigations, date=date, allow_open_ended=allow_open_ended
        )
        if investigation is None:
            return dict()  # no valid investigation found

        # Normalize information
        for key in ["parameters", "visitId"]:
            investigation.pop(key, None)
        ninfo = dict()
        ninfo["proposal"] = investigation.pop("name", None)
        ninfo["beamline"] = investigation.pop("instrument", dict()).get("name", None)
        ninfo.update(investigation)
        ninfo["e-logbook"] = (
            f"https://data.esrf.fr/investigation/{investigation['id']}/logbook"
        )
        ninfo["data portal"] = (
            f"https://data.esrf.fr/investigation/{investigation['id']}/datasets"
        )

        self.__investigation_info[investigation_key] = ninfo
        return ninfo

    def _investigation_id(
        self,
        beamline: str,
        proposal: str,
        date: Optional[Union[datetime.datetime, datetime.date]] = None,
        allow_open_ended: bool = True,
        timeout: Optional[float] = None,
    ) -> Optional[int]:
        info = self.investigation_info(
            beamline=beamline,
            proposal=proposal,
            date=date,
            allow_open_ended=allow_open_ended,
            timeout=timeout,
        )
        if info is None:
            return None
        return info.get("id", None)

    def registered_dataset_ids(
        self,
        beamline: str,
        proposal: str,
        date: Optional[Union[datetime.datetime, datetime.date]] = None,
        allow_open_ended: bool = True,
        timeout: Optional[float] = None,
    ) -> Optional[List[DatasetId]]:
        investigation_id = self._investigation_id(
            beamline=beamline,
            proposal=proposal,
            date=date,
            allow_open_ended=allow_open_ended,
            timeout=timeout,
        )
        if investigation_id is None:
            return None
        url = self._dataset_url.format(investigation_id=investigation_id)
        datasets = self._get_with_response_parsing(url, timeout=timeout)
        if datasets is None:
            return None  # not available at the moment
        return [self._icat_dataset_to_datasetid(dataset) for dataset in datasets]

    def registered_datasets(
        self,
        beamline: str,
        proposal: str,
        date: Optional[Union[datetime.datetime, datetime.date]] = None,
        allow_open_ended: bool = True,
        timeout: Optional[float] = None,
    ) -> Optional[List[Dataset]]:
        investigation_id = self._investigation_id(
            beamline=beamline,
            proposal=proposal,
            date=date,
            allow_open_ended=allow_open_ended,
            timeout=timeout,
        )
        if investigation_id is None:
            return None
        url = self._dataset_url.format(investigation_id=investigation_id)
        datasets = self._get_with_response_parsing(url, timeout=timeout)
        if datasets is None:
            return None  # not available at the moment
        return [self._icat_dataset_to_dataset(dataset) for dataset in datasets]

    @staticmethod
    def _get_dataset_parameter_value_by_name(
        dataset: dict, parameter_name: str
    ) -> Optional[str]:
        if "parameters" in dataset:
            for p in dataset["parameters"]:
                if p["name"] == parameter_name:
                    return p["value"]

    @classmethod
    def _icat_dataset_to_datasetid(cls, dataset: dict) -> DatasetId:
        location = dataset["location"]
        location, name = os.path.split(location)
        while location and not name:
            location, name = os.path.split(location)
        return DatasetId(name=name, path=dataset["location"])

    @classmethod
    def _icat_dataset_to_dataset(cls, dataset: dict) -> Dataset:
        dataset_id = cls._icat_dataset_to_datasetid(dataset)
        file_count = cls._get_dataset_parameter_value_by_name(dataset, "__fileCount")
        dataset_metadata = DatasetMetadata(
            file_count=int(0 if file_count is None else file_count)
        )
        return Dataset(
            dataset_id=dataset_id,
            icat_dataset_id=dataset["id"],
            dataset_metadata=dataset_metadata,
        )


def _select_investigation(
    investigations: List[dict],
    date: Optional[Union[datetime.datetime, datetime.date]] = None,
    allow_open_ended: bool = True,
) -> Optional[dict]:
    """When `date` is not provided we take it to be "now".

    This method returns the last investigation that contains
    the date or has a start/end closest to the date. The
    investigations are ordered from first to last created.

    Optionally all open-ended investigations can be ignored.
    Open-ended investigations have a start date but no end date.
    These investigations are created by sending dataset or
    investigation messages with start dates 48h outside any
    official investigation.

    :param investigations: unsorted list of investigations
    :param date: possibly not fully deterministic date
    :param allow_open_ended: validation option
    :returns: matching investigation according to the selection rules
    """
    investigations = _filter_valid_investigations(
        investigations, allow_open_ended=allow_open_ended
    )
    if not investigations:
        return
    if len(investigations) == 1:
        return investigations[0]

    # Sort investigations by order of importance: database creation time
    investigations = sorted(
        investigations, key=lambda investigation: investigation["id"]
    )

    date = _deterministic_date(date, investigations)
    return _select_investigation_from_sorted_list(investigations, date)


def _filter_valid_investigations(
    investigations: List[dict], allow_open_ended: bool = True
) -> List[dict]:
    """Filter out investigations with invalid time slots.

    :param investigations: unsorted list of investigations
    :param allow_open_ended: validation option
    :returns: list of valid investigations (preserve order)
    """
    if allow_open_ended:
        return [
            investigation
            for investigation in investigations
            if investigation.get("startDate")
        ]
    return [
        investigation
        for investigation in investigations
        if investigation.get("startDate") and investigation.get("endDate")
    ]


def _deterministic_date(
    date: Optional[Union[datetime.datetime, datetime.date]], investigations: List[dict]
) -> datetime.datetime:
    """The date might

    * not exist: take now
    * have no time zone: add local timezone
    * be a day without time: select time from the investigations or the default 8 a.m.

    The resulting date is fully deterministic.

    :param date: possibly not fully deterministic date
    :param investigations: sorted by order of importance
    :returns: fully deterministic date
    """
    if date is None:
        return datetime.datetime.now().astimezone()

    if isinstance(date, datetime.datetime):
        return date.astimezone()

    # Get the time from the investigations that start on the same day
    start_dates = [
        _tz_aware_fromisoformat(investigation["startDate"])
        for investigation in investigations
    ]
    is_start_date = [dt.date() == date for dt in start_dates]
    last_is_start_date = _last_where_true(is_start_date)
    if last_is_start_date is None:
        # No investigation starts on the same day
        start_time = _DEFAULT_START_TIME
    else:
        # Last investigation that starts on the same day
        start_time = start_dates[last_is_start_date].time()

    date = datetime.datetime.combine(date, start_time)
    return date.astimezone()


def _select_investigation_from_sorted_list(
    investigations: List[dict], date: datetime.datetime
) -> dict:
    """
    :param investigations: sorted by order of importance
    :param date: point in time for which we want to find the corresponding investigation
    :returns: matching investigation according to the selection rules
    """
    # Seconds between date and start/end of each investigation
    n = len(investigations)
    startdiff = numpy.zeros(n)
    enddiff = numpy.full(n, numpy.inf)
    for i, investigation in enumerate(investigations):
        startdate = _tz_aware_fromisoformat(investigation["startDate"])
        startdiff[i] = (date - startdate).total_seconds()
        enddate = investigation.get("endDate")
        if enddate is not None:
            enddate = _tz_aware_fromisoformat(enddate)
            enddiff[i] = (enddate - date).total_seconds()

    # Close open-ended investigation when the next investigation starts (if any)
    closed_ended = numpy.isfinite(enddiff)
    starttime_order = numpy.argsort(-startdiff)
    for i in range(n - 1):
        idx = starttime_order[i]
        if not closed_ended[idx]:
            idx_next = starttime_order[i + 1]
            enddiff[idx] = -startdiff[idx_next]

    # Return the last closed investigation which contains the date (if any)
    contains_date = (startdiff >= 0) & (enddiff >= 0)
    contains_date_and_closed = contains_date & closed_ended
    idx = _last_where_true(contains_date_and_closed)
    if idx is not None:
        return investigations[idx]

    # Return the last investigation which contains the date (if any)
    idx = _last_where_true(contains_date)
    if idx is not None:
        return investigations[idx]

    # Return the last investigation with the closest start or end date
    startdiff = numpy.abs(startdiff)
    enddiff = numpy.abs(enddiff)
    istart = numpy.argmin(startdiff)
    iend = numpy.argmin(enddiff)
    min_startdiff = startdiff[istart]
    min_enddiff = enddiff[iend]
    if min_startdiff < min_enddiff:
        return investigations[istart]
    if min_startdiff > min_enddiff:
        return investigations[iend]
    return investigations[max(istart, iend)]


def _tz_aware_fromisoformat(date: str) -> datetime.datetime:
    return datetime_utils.deserialize_datetime(date).astimezone()


def _last_where_true(bool_arr: Sequence[bool]) -> Optional[int]:
    indices = numpy.argwhere(bool_arr)
    if indices.size:
        return indices[-1][0]



import html

_SELF_CLOSING = {"br", "img", "hr", "meta", "link", "input"}


def _normalize_attr_name(name):
    if name.endswith("_"):
        name = name[:-1]
    if name.startswith("_"):
        name = name[1:]
    return name


def _format_attrs(attrs):
    if not attrs:
        return ""
    parts = []
    for key, value in attrs.items():
        if value is None:
            continue
        key = _normalize_attr_name(key)
        value = html.escape(str(value), quote=True)
        parts.append(f' {key}="{value}"')
    return "".join(parts)


class _Tag:
    def __init__(self, page, name):
        self._page = page
        self._name = name

    def __call__(self, *content, **attrs):
        attr_text = _format_attrs(attrs)
        if self._name in _SELF_CLOSING:
            self._page._append(f"<{self._name}{attr_text} />")
            return None
        if content:
            inner = "".join(str(part) for part in content)
            self._page._append(f"<{self._name}{attr_text}>{inner}</{self._name}>")
            return None
        self._page._append(f"<{self._name}{attr_text}>")
        return None

    def close(self):
        if self._name in _SELF_CLOSING:
            return None
        self._page._append(f"</{self._name}>")
        return None


class HtmlPage:
    def __init__(self, mode=None):
        self._parts = []
        self._title = ""
        self._footer = ""
        self._mode = mode
        self._init_called = False

    def _append(self, text):
        self._parts.append(text)

    def init(self, title="", footer=""):
        self._title = title
        self._footer = footer
        self._init_called = True

    def escape(self, text):
        return html.escape(str(text), quote=True)

    def __getattr__(self, name):
        if name.startswith("_"):
            raise AttributeError(name)
        tag = _Tag(self, name)
        self.__dict__[name] = tag
        return tag

    def __str__(self):
        title = html.escape(self._title, quote=True) if self._title else ""
        footer = html.escape(self._footer, quote=True) if self._footer else ""
        parts = [
            "<!DOCTYPE html>",
            "<html>",
            "<head>",
            '<meta charset="utf-8" />',
        ]
        if title:
            parts.append(f"<title>{title}</title>")
        parts.extend(["</head>", "<body>"])
        parts.extend(self._parts)
        if footer:
            parts.append(f'<div class="footer">{footer}</div>')
        parts.extend(["</body>", "</html>"])
        return "\n".join(parts)



import base64

import pytest


def test_elogbook_message_wrong_category(elogbook_client):
    client, messages = elogbook_client
    with pytest.raises(ValueError):
        client.send_message(
            "mycontent",
            message_type="wrongcategory",
            beamline="id00",
            proposal="hg123",
            dataset="datasetname",
        )
    assert messages.empty()


def test_elogbook_message(elogbook_client):
    client, messages = elogbook_client
    client.send_message(
        "mycontent",
        message_type="comment",
        beamline="id00",
        proposal="hg123",
        dataset="datasetname",
    )
    message = messages.get(timeout=10)
    assert message.pop("apikey")
    assert message.pop("creationDate")
    assert message.pop("machine")
    assert message.pop("software").startswith("pyicat-plus")
    expected = {
        "type": "annotation",
        "datasetName": "datasetname",
        "category": "comment",
        "content": [{"format": "plainText", "text": "mycontent"}],
        "investigation": "hg123",
        "instrument": "id00",
    }
    assert message == expected
    assert messages.empty()


@pytest.mark.parametrize(
    "editable", [None, True, False], ids=["noneeditable", "editable", "noteditable"]
)
@pytest.mark.parametrize(
    "formatted", [None, True, False], ids=["noneformatted", "formatted", "notformatted"]
)
@pytest.mark.parametrize(
    "message_type", ["debug", "info", "warning", "error", "comment", "command"]
)
@pytest.mark.parametrize("mimetype", ["text/plain", "text/html"])
def test_elogbook_message_options(
    elogbook_client, editable, formatted, message_type, mimetype
):
    client, messages = elogbook_client

    # See https://confluence.esrf.fr/display/DATAPOLWK/Electronic+Logbook#Summary
    if message_type == "comment":
        if formatted is True and editable in (False, None):
            msg_type = "notification"
        elif editable is False and formatted in (True, None):
            msg_type = "notification"
        else:
            msg_type = "annotation"
    else:
        msg_type = "notification"

    has_warning = editable == formatted and formatted is not None
    if message_type != "comment":
        has_warning |= editable is True or formatted is False

    # Check message sending
    if has_warning:
        with pytest.warns(UserWarning):
            client.send_message(
                "mycontent",
                message_type=message_type,
                beamline="id00",
                proposal="hg123",
                editable=editable,
                formatted=formatted,
                mimetype=mimetype,
            )
    else:
        client.send_message(
            "mycontent",
            message_type=message_type,
            beamline="id00",
            proposal="hg123",
            editable=editable,
            formatted=formatted,
            mimetype=mimetype,
        )
    message = messages.get(timeout=10)
    assert messages.empty()

    # Check message content
    if message_type == "command":
        category = "commandLine"
    elif message_type == "warning":
        category = "error"
    else:
        category = message_type
    assert message["category"] == category

    assert message["type"] == msg_type

    if mimetype == "text/plain":
        format = "plainText"
    else:
        format = "html"
    assert message["content"][0]["format"] == format


def test_ebs_elogbook_message(elogbook_ebs_client):
    client, messages = elogbook_ebs_client

    client.send_message("mycontent", message_type="error")
    message = messages.get(timeout=10)
    assert message.pop("apikey")
    assert message.pop("creationDate")
    assert message.pop("machine")
    expected = {
        "type": "broadcast",
        "source": "ebs",
        "category": "error",
        "software": "testsoft",
        "tag": [{"name": "testtag"}, {"name": "machine"}],
        "content": [{"format": "plainText", "text": "mycontent"}],
    }
    assert message == expected

    client.send_message(
        "mycontent", message_type="comment", tags=["commenttag"], software="mysoft"
    )
    message = messages.get(timeout=10)
    assert message.pop("apikey")
    assert message.pop("creationDate")
    assert message.pop("machine")
    expected = {
        "type": "broadcast",
        "source": "ebs",
        "category": "comment",
        "software": "mysoft",
        "tag": [{"name": "testtag"}, {"name": "machine"}, {"name": "commenttag"}],
        "content": [{"format": "plainText", "text": "mycontent"}],
    }
    assert message == expected

    assert messages.empty()


def test_elogbook_message_beamline_only(elogbook_client):
    client, messages = elogbook_client
    client.send_message("mycontent", message_type="comment", beamline="id00")
    message = messages.get(timeout=10)
    assert message.pop("apikey")
    assert message.pop("creationDate")
    assert message.pop("machine")
    assert message.pop("software").startswith("pyicat-plus")
    expected = {
        "type": "annotation",
        "category": "comment",
        "content": [{"format": "plainText", "text": "mycontent"}],
        "instrument": "id00",
    }
    assert message == expected
    assert messages.empty()


def test_elogbook_data(elogbook_client):
    client, messages = elogbook_client
    client.send_binary_data(
        b"123", mimetype="application/octet-stream", beamline="id00", proposal="hg123"
    )
    message = messages.get(timeout=10)
    assert message.pop("apikey")
    assert message.pop("creationDate")
    assert message.pop("machine")
    assert message.pop("software").startswith("pyicat-plus")
    data = message.pop("base64")
    data = data.replace("data:application/octet-stream;base64,", "")
    assert base64.b64decode(data.encode()) == b"123"
    expected = {"investigation": "hg123", "instrument": "id00"}
    assert message == expected
    assert messages.empty()


def test_elogbook_data_beamline_only(elogbook_client):
    client, messages = elogbook_client
    client.send_binary_data(
        b"123", mimetype="application/octet-stream", beamline="id00"
    )
    message = messages.get(timeout=10)
    assert message.pop("apikey")
    assert message.pop("creationDate")
    assert message.pop("machine")
    assert message.pop("software").startswith("pyicat-plus")
    data = message.pop("base64")
    data = data.replace("data:application/octet-stream;base64,", "")
    assert base64.b64decode(data.encode()) == b"123"
    expected = {"instrument": "id00"}
    assert message == expected
    assert messages.empty()



import datetime
from dataclasses import is_dataclass
from typing import Any
from typing import Dict
from typing import Optional

from xsdata.formats.dataclass.serializers import XmlSerializer
from xsdata.formats.dataclass.serializers.config import SerializerConfig


def serialize_message(xsdata_model: Any, indent: Optional[str] = None) -> bytes:
    """Serialize message with XSD validation."""
    if not is_dataclass(xsdata_model):
        raise TypeError("Expected a dataclass")
    config = SerializerConfig(indent=indent, xml_declaration=True)
    serializer = XmlSerializer(config=config)
    xml_string = serializer.render(
        xsdata_model, ns_map={"tns": "http://www.esrf.fr/icat"}
    )
    if indent is None:
        xml_string = xml_string.replace("\n", "")
    return xml_string.encode("utf-8")


def serialize_root(data: Dict[str, Any]) -> Dict[str, Any]:
    """XSD validation currently does not validate values so we do it ourselves."""
    result = {}
    for name, value in data.items():
        if isinstance(value, datetime.datetime):
            value = value.isoformat()
        result[name] = value
    return result



import logging
import sys
from typing import Optional


def basic_config(
    logger: Optional[logging.Logger] = None,
    level: Optional[int] = None,
    format: Optional[str] = None,
) -> None:
    """
    :param logger: root logger when not provided
    :param level: logger log level
    :param str format:
    """
    if logger is None:
        logger = logging.getLogger()
    if level is not None:
        logger.setLevel(level)

    threshold_level = logging.WARNING

    class StdOutFilter(logging.Filter):
        def filter(self, record):
            return record.levelno < threshold_level

    class StdErrFilter(logging.Filter):
        def filter(self, record):
            return record.levelno >= threshold_level

    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.addFilter(StdOutFilter())
    stdout_handler.setLevel(logging.DEBUG)
    logger.addHandler(stdout_handler)

    stderr_handler = logging.StreamHandler(sys.stderr)
    stderr_handler.addFilter(StdErrFilter())
    stderr_handler.setLevel(threshold_level)
    logger.addHandler(stderr_handler)

    if format:
        formatter = logging.Formatter(format)
        stdout_handler.setFormatter(formatter)
        stderr_handler.setFormatter(formatter)



import datetime

import pytest

from ... import errors
from ..._icat_messaging.serializers.dataset_parameters.fallback import (
    fallback_serialization,
)
from ..._icat_messaging.utils import datetime_utils


@pytest.mark.parametrize(
    "input_data, expected",
    [
        ({}, {}),
        ({"key": None}, {}),
        ({"key": "string"}, {"key": "string"}),
        ({"key": b"string"}, {"key": "string"}),
        ({"key": 123.456}, {"key": "123.456"}),
        ({"key": []}, {"key": ""}),
        ({"key": [None]}, {"key": ""}),
        ({"key": [1, 2]}, {"key": "1,2"}),
        ({"key": [[1, 2], [3, 4]]}, {"key": "1,2 3,4"}),
    ],
    ids=[
        "empty-dict",
        "none-value",
        "string",
        "bytes",
        "float",
        "empty-list",
        "list-with-none",
        "flat-list",
        "nested-list",
    ],
)
def test_icat_serialize_fallback_valid_data(input_data, expected):
    assert fallback_serialization(input_data) == expected


def test_icat_serialize_fallback_datetime():
    now = datetime.datetime.now().astimezone()
    actual = fallback_serialization({"key": now})
    expected = {"key": datetime_utils.with_timezone_zulu_format(now)}
    assert actual == expected


@pytest.mark.parametrize(
    "invalid_value",
    [[{}], [[[1, 2], [3, 4]], [[5, 6], [7, 8]]], (7.3, "keV")],
    ids=["list-containing-dict", "3d-nested-list", "mixed-types"],
)
def test_icat_serialize_fallback_invalid_data(invalid_value):
    with pytest.raises(
        errors.IcatMetadataSerializationError, match="Invalid ICAT dataset parameter"
    ):
        fallback_serialization({"key": invalid_value})


def test_icat_serialize_fallback_boolean():
    actual = fallback_serialization({"key": True})
    expected = {"key": "true"}
    assert actual == expected

    actual = fallback_serialization({"key": False})
    expected = {"key": "false"}
    assert actual == expected



import base64
import html
import json
import os
import tempfile
import time
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import PIL.Image

from ..utils import html_utils

REPORT_VERSION = 1.0


class DracReportBuilder(object):
    """Build a json and/or html file compatible with Drac dataset reports.

    Example:
        report = DracReportBuilder(
            dataset_type="Characterisation",
            dataset_title="Characterisation Results",
        )
        report.add_info("This is a demo info item.")
        report.add_table(
            title="Sample Table",
            columns=["Name", "Value"],
            data=[["Alpha", 1], ["Beta", 2]],
        )
        report.add_image("/path/to/image.png", image_title="Preview")
        report.render_json("report.json")
    """

    def __init__(self, dataset_type: str, dataset_title: str = "") -> None:
        self.image_list: Optional[Dict[str, Any]] = None
        self.dict_report: Dict[str, Any] = {}
        self.dict_report["version"] = REPORT_VERSION
        self.dict_report["type"] = dataset_type
        self.dict_report["title"] = dataset_title
        self.dict_report["items"] = []

    def set_title(self, dataset_title: str) -> None:
        self.dict_report["title"] = dataset_title

    def add_info(self, info_text: str) -> None:
        self.dict_report["items"].append({"type": "info", "value": info_text})

    def add_warning(self, info_text: str) -> None:
        self.dict_report["items"].append({"type": "warning", "value": info_text})

    def start_image_list(self) -> None:
        self.image_list = {}
        self.image_list["type"] = "images"
        self.image_list["items"] = []

    def end_image_list(self) -> None:
        self.dict_report["items"].append(self.image_list)
        self.image_list = None

    def _create_image_item(
        self,
        path_to_image: str,
        image_title: str,
        path_to_thumbnail_image: Optional[str] = None,
        thumbnail_height: Optional[int] = None,
        thumbnail_width: Optional[int] = None,
    ) -> Dict[str, Any]:
        item: Dict[str, Any] = {}
        item["type"] = "image"
        item["suffix"] = path_to_image.split(".")[-1]
        item["title"] = image_title
        with PIL.Image.open(path_to_image) as image:
            item["xsize"] = image.size[0]
            item["ysize"] = image.size[1]
        with open(path_to_image, "rb") as image_file:
            item["value"] = base64.b64encode(image_file.read()).decode("utf-8")
        if path_to_thumbnail_image is None:
            if thumbnail_height is not None and thumbnail_width is not None:
                item["thumbnailSuffix"] = path_to_image.split(".")[-1]
                item["thumbnailXsize"] = thumbnail_height
                item["thumbnailYsize"] = thumbnail_width
                with open(path_to_image, "rb") as image_file:
                    item["thumbnailValue"] = base64.b64encode(image_file.read())
        else:
            item["thumbnailSuffix"] = path_to_thumbnail_image.split(".")[-1]
            with PIL.Image.open(path_to_thumbnail_image) as thumbnail_im:
                item["thumbnailXsize"] = thumbnail_im.size[0]
                item["thumbnailYsize"] = thumbnail_im.size[1]
            with open(path_to_thumbnail_image, "rb") as thumbnail_file:
                item["thumbnailValue"] = base64.b64encode(thumbnail_file.read()).decode(
                    "utf-8"
                )
        return item

    def add_image(
        self,
        path_to_image: str,
        image_title: str = "",
        path_to_thumbnail_image: Optional[str] = None,
        thumbnail_height: Optional[int] = None,
        thumbnail_width: Optional[int] = None,
    ) -> None:
        if self.image_list is None:
            self.dict_report["items"].append(
                self._create_image_item(
                    path_to_image,
                    image_title,
                    path_to_thumbnail_image,
                    thumbnail_height,
                    thumbnail_width,
                )
            )
        else:
            self.image_list["items"].append(
                self._create_image_item(
                    path_to_image,
                    image_title,
                    path_to_thumbnail_image,
                    thumbnail_height,
                    thumbnail_width,
                )
            )

    def add_table(
        self,
        title: str,
        columns: List[str],
        data: List[List[Any]],
        orientation: str = "horizontal",
    ) -> None:
        item: Dict[str, Any] = {}
        item["type"] = "table"
        item["title"] = title
        item["columns"] = columns
        item["data"] = data
        item["orientation"] = orientation
        self.dict_report["items"].append(item)

    def add_log_file(self, title: str, link_text: str, path_to_log_file: str) -> None:
        item: Dict[str, Any] = {}
        item["type"] = "logFile"
        item["title"] = title
        item["link_text"] = link_text
        with open(path_to_log_file, "r", encoding="utf-8", errors="ignore") as log_file:
            item["logText"] = log_file.read()
        self.dict_report["items"].append(item)

    def get_dict_report(self) -> Dict[str, Any]:
        return self.dict_report

    def render_json(self, path_to_json_dir: str) -> str:
        path_to_json_file = os.path.join(path_to_json_dir, "report.json")
        with open(path_to_json_file, "w") as json_file:
            json_file.write(json.dumps(self.dict_report, indent=4))
        return path_to_json_file

    def escape_characters(self, str_value: str) -> str:
        str_value = html.escape(str_value)
        str_value = str_value.replace(str("Å"), "&Aring;")
        str_value = str_value.replace(str("°"), "&deg;")
        str_value = str_value.replace("\n", "<br>")
        return str_value

    def render_html(
        self, path_to_html_dir: str, name_of_index_file: str = "index.html"
    ) -> str:
        page = html_utils.HtmlPage(mode="loose_html")
        page.init(
            title=self.dict_report["title"], footer="Generated on %s" % time.asctime()
        )
        page.div(align_="LEFT")
        page.h1()
        page.strong(self.dict_report["title"])
        page.h1.close()
        page.div.close()
        for item in self.dict_report["items"]:
            if "value" in item:
                item_value = item["value"]
            if "title" in item:
                item_title = self.escape_characters(item["title"])
            if item["type"] == "info":
                page.p(item_value)
            if item["type"] == "warning":
                page.font(_color="red", size="+1")
                page.p()
                page.strong(item_value)
                page.p.close()
                page.font.close()
            elif item["type"] == "image":
                self._render_image(page, item, path_to_html_dir)
                page.br()
                page.p(item_title)
            elif item["type"] == "images":
                page.table()
                page.tr(align_="CENTER")
                for item in item["items"]:
                    item_title = self.escape_characters(item["title"])
                    page.td()
                    page.table()
                    page.tr()
                    page.td()
                    self._render_image(page, item, path_to_html_dir)
                    page.td.close()
                    page.tr.close()
                    page.tr(align_="CENTER")
                    page.td(item_title)
                    page.tr.close()
                    page.table.close()
                    page.td.close()
                page.tr.close()
                page.table.close()
                page.br()
            elif item["type"] == "table":
                title_bg_colour = "#99c2ff"
                column_title_bg_colour = "#ffffff"
                data_bg_colour = "#e6f0ff"
                page.table(
                    border_="1",
                    cellpadding_="2",
                    width="100%",
                    style_=(
                        "border: 1px solid black; border-collapse: collapse; margin:"
                        " 0px; font-size: 12px"
                    ),
                )
                page.tr(align_="LEFT")
                # page.strong(item_title)
                page.th(
                    item_title,
                    bgcolor_=title_bg_colour,
                    align_="LEFT",
                    style_=(
                        "border: 1px solid black; border-collapse: collapse; padding:"
                        " 5px;"
                    ),
                    colspan_=str(len(item["columns"])),
                )
                page.tr.close()
                if "orientation" in item and item["orientation"] == "vertical":
                    for index1 in range(len(item["columns"])):
                        item_column = self.escape_characters(item["columns"][index1])
                        page.tr(align_="LEFT")
                        page.th(
                            item_column,
                            bgcolor_=column_title_bg_colour,
                            align_="LEFT",
                            style_=(
                                "border: 1px solid black; border-collapse: collapse;"
                                " padding: 5px;"
                            ),
                        )
                        for index2 in range(len(item["data"])):
                            item_data = self.escape_characters(
                                str(item["data"][index2][index1])
                            )
                            page.th(
                                item_data,
                                bgcolor_=data_bg_colour,
                                style_="padding: 5px;",
                            )
                        page.tr.close()
                else:
                    page.tr(
                        align_="LEFT",
                        bgcolor_=column_title_bg_colour,
                        style_=(
                            "border: 1px solid black; border-collapse: collapse;"
                            " padding: 5px;"
                        ),
                    )
                    for column in item["columns"]:
                        item_column = self.escape_characters(column)
                        page.th(
                            item_column,
                            style_=(
                                "border: 1px solid black; border-collapse: collapse;"
                                " padding: 5px;"
                            ),
                        )
                    page.tr.close()
                    for list_row in item["data"]:
                        page.tr(
                            align_="LEFT",
                            bgcolor_=data_bg_colour,
                            style_="border-collapse: collapse; padding: 5px;",
                        )
                        for cell in list_row:
                            item_cell = self.escape_characters(str(cell))
                            page.th(
                                item_cell,
                                style_=(
                                    "border: 1px solid black; border-collapse:"
                                    " collapse; padding: 5px;"
                                ),
                            )
                        page.tr.close()
                page.table.close()
                page.br()
            elif item["type"] == "logFile":
                path_to_log_html = os.path.join(path_to_html_dir, item_title + ".html")
                if os.path.exists(path_to_log_html):
                    fd, path_to_log_html = tempfile.mkstemp(
                        suffix=".html",
                        prefix=item_title.replace(" ", "_") + "_",
                        dir=path_to_html_dir,
                    )
                    os.close(fd)
                page_log_html = html_utils.HtmlPage()
                page_log_html.h1(item_title)
                page_log_html.pre(page.escape(item["logText"]))
                with open(path_to_log_html, "w") as log_html_file:
                    log_html_file.write(str(page_log_html))
                os.chmod(path_to_log_html, 0o644)
                page.p()
                page.a(item["link_text"], href_=os.path.basename(path_to_log_html))
                page.p.close()
        html = str(page)
        page_path = os.path.join(path_to_html_dir, name_of_index_file)
        with open(page_path, "w") as file_page:
            file_page.write(html)
        return page_path

    def _render_image(
        self, page: Any, item: Dict[str, Any], path_to_html_dir: str
    ) -> None:
        image_name = item["title"].replace(" ", "_")
        path_to_image = os.path.join(
            path_to_html_dir, "{0}.{1}".format(image_name, item["suffix"])
        )
        if os.path.exists(path_to_image):
            fd, path_to_image = tempfile.mkstemp(
                suffix="." + item["suffix"],
                prefix=image_name + "_",
                dir=path_to_html_dir,
            )
            os.close(fd)
        with open(path_to_image, "wb") as image_file:
            image_file.write(base64.b64decode(item["value"]))
        os.chmod(path_to_image, 0o644)
        if "thumbnailValue" in item:
            thumbnail_image_name = image_name + "_thumbnail"
            path_to_thumbnail_image = os.path.join(
                path_to_html_dir, "{0}.{1}".format(thumbnail_image_name, item["suffix"])
            )
            if os.path.exists(path_to_thumbnail_image):
                fd, path_to_thumbnail_image = tempfile.mkstemp(
                    suffix="." + item["suffix"],
                    prefix=thumbnail_image_name + "_",
                    dir=path_to_html_dir,
                )
                os.close(fd)
            with open(path_to_thumbnail_image, "wb") as thumbnail_file:
                thumbnail_file.write(base64.b64decode(item["thumbnailValue"]))
            os.chmod(path_to_thumbnail_image, 0o644)
            page_reference_image = html_utils.HtmlPage(mode="loose_html")
            page_reference_image.init(
                title=image_name, footer="Generated on %s" % time.asctime()
            )
            page_reference_image.h1(image_name)
            page_reference_image.br()
            page_reference_image.img(
                src=os.path.basename(path_to_image),
                title=image_name,
                width=item["xsize"],
                height=item["ysize"],
            )
            page_reference_image.br()
            path_page_reference_image = os.path.join(
                path_to_html_dir, "{0}.html".format(image_name)
            )
            with open(path_page_reference_image, "w") as file_page:
                file_page.write(str(page_reference_image))
            page.a(href=os.path.basename(path_page_reference_image))
            page.img(
                src=os.path.basename(path_to_thumbnail_image),
                title=image_name,
                width=item["thumbnailXsize"],
                height=item["thumbnailYsize"],
            )
            page.a.close()
        else:
            page.img(
                src=os.path.basename(path_to_image),
                title=item["title"],
                width=item["xsize"],
                height=item["ysize"],
            )



from importlib.resources import as_file
from importlib.resources import files
from typing import BinaryIO

_ENTERED_CONTEXT_MANAGERS = []


def resource_filename(package: str, resource: str) -> str:
    """
    Replacement for pkg_resources.resource_filename
    Returns a real filesystem path (even if package is zipped)
    """
    ref = files(package) / resource
    cm = as_file(ref)
    path = cm.__enter__()
    _ENTERED_CONTEXT_MANAGERS.append(cm)
    return str(path)


def resource_stream(package: str, resource: str) -> BinaryIO:
    """
    Replacement for pkg_resources.resource_stream
    Returns a file-like object (rb)
    """
    return (files(package) / resource).open("rb")


def cleanup() -> None:
    """
    Close any open context managers
    """
    global _ENTERED_CONTEXT_MANAGERS
    for cm in _ENTERED_CONTEXT_MANAGERS:
        try:
            cm.__exit__(None, None, None)
        except Exception:
            pass
    _ENTERED_CONTEXT_MANAGERS = []



import datetime
import os
import socket
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Optional
from typing import Union

from xsdata.formats.dataclass.parsers import DictDecoder

from ... import errors
from ..models.xsd.dataset import Dataset
from ..utils import constants
from . import xml
from .dataset_parameters.main import serialize_dataset_parameters


def serialize_dataset_message(
    beamline: str,
    proposal: str,
    dataset: str,
    path: Union[str, Path],
    metadata: Optional[Dict[str, Any]] = None,
    strict: bool = False,
    indent: Optional[str] = None,
) -> bytes:
    """
    :raises IcatMetadataSerializationError: If metadata serialization fails.
    :raises IcatMetadataValidationError: If ``strict=True`` and the metadata is invalid.
    :emits IcatMetadataValidationWarning: If ``strict=False`` and the metadata is invalid.
    :emits IcatMetadataDeprecatedWarning: If metadata is deprecated.
    """
    # Early raise missing root keys
    if not beamline:
        raise AssertionError("ICAT requires the beamline name")
    if not proposal:
        raise AssertionError("ICAT requires the proposal name")
    if not dataset:
        raise AssertionError("ICAT requires the dataset name")
    if not path:
        raise AssertionError("ICAT requires the dataset path")

    # Parameters
    if metadata:
        parameters = metadata.copy()
    else:
        parameters = {}

    # Required parameters
    if "Sample_name" not in parameters:
        raise AssertionError("ICAT metadata field 'Sample_name' is missing")

    # Default parameters
    now = datetime.datetime.now().astimezone()
    _ = parameters.setdefault("startDate", now)
    _ = parameters.setdefault("endDate", now)
    _ = parameters.setdefault("machine", socket.getfqdn())
    _ = parameters.setdefault("software", "pyicat-plus_v" + constants.release)
    _ = parameters.setdefault("complete", True)

    # Parameters that are provided as arguments
    _set_dataset_parameter(parameters, "datasetName", "dataset", dataset, strict)
    _set_dataset_parameter(
        parameters, "location", "path", path, strict, normalize=os.path.normpath
    )
    _set_dataset_parameter(parameters, "proposal", "proposal", proposal, strict)

    # Normalize parameters
    input_datasets = parameters.get("input_datasets", None)
    if input_datasets:
        parameters["input_datasets"] = [
            os.path.normpath(input_dataset) for input_dataset in input_datasets
        ]

    # XML message root
    root = {
        # Root element only
        "instrument": beamline,
        # Both root attribute and parameter value
        "complete": parameters["complete"],
        # Both root element and parameter value
        "investigation": parameters["proposal"],
        "name": parameters["datasetName"],
        "location": parameters["location"],
        "startDate": parameters["startDate"],
        "endDate": parameters["endDate"],
        # Both sample element and parameter value
        "sample": {
            "name": parameters["Sample_name"],
        },
    }

    # Serialize before XSD serialization
    root = xml.serialize_root(root)
    parameters = serialize_dataset_parameters(parameters, strict=strict)

    # Add parameters to root
    if parameters:
        root["parameter"] = []
    for name, value in parameters.items():
        root["parameter"].append({"name": name, "value": value})

    # Trick XSD validation (it currently does not validate values, only keys)
    root["sample"]["type"] = None

    # Serialize with XSD validation
    return raw_serialize_dataset_message(root, indent=indent)


def raw_serialize_dataset_message(
    data: Dict[str, Any], indent: Optional[str] = None
) -> bytes:
    """Serialize message with XSD 'dataset' schema."""
    model = DictDecoder().decode(data, Dataset)
    return xml.serialize_message(model, indent=indent)


def _set_dataset_parameter(
    parameters: Dict[str, Any],
    key: str,
    argument_name: str,
    argument_value: Any,
    strict: bool,
    normalize: Optional[callable] = None,
) -> None:
    """
    Normalize and set a dataset parameter to the value provided by a serialization argument.

    If the parameter is already set, overwrite it and emit a warning
    (or raise exception when strict=True) if they are not the same.
    """
    parameter_value = parameters.pop(key, None)

    if normalize:
        argument_value = normalize(argument_value)
        if parameter_value is not None:
            parameter_value = normalize(parameter_value)

    parameters[key] = argument_value

    if parameter_value is not None and parameter_value != argument_value:
        message = f"ICAT dataset metadata {key}={parameter_value} is not equal "
        f"to serialization argument {argument_name}={argument_value}. Use the argument value."
        errors.invalid_warn_or_raise(message, strict=strict)



import datetime
import json
import logging
import os
import re
import time
from contextlib import contextmanager
from pathlib import Path
from pprint import pprint
from typing import Callable
from typing import Generator
from typing import Optional
from typing import Tuple

import h5py

from .._icat_messaging.utils import datetime_utils
from ..apps import sync_raw
from ..client.main import IcatClient
from ..utils import path_utils
from ..utils import sync_types


def test_unregistered_datasets(tmp_path: Path, icat_main_client, icat_data_dir: Path):
    client, messages = icat_main_client
    root_dir = tmp_path

    exp_session_test = _init_unregistered_experiment(
        client, messages, icat_data_dir, root_dir, "hg333", "id99"
    )

    args = client, "hg333", "id99", "20231019"
    kwargs = {"root_dir": str(root_dir), "raw_data_format": "esrfv3"}

    exp_session = sync_raw.sync_session(*args, **kwargs)
    assert len(exp_session.datasets["registered"]) == 0
    assert len(exp_session.datasets["unregistered"]) == 6

    if exp_session != exp_session_test:
        assert exp_session.as_dict() == exp_session_test.as_dict()

    first_datasets = exp_session_test.datasets["unregistered"][:3]
    last_datasets = exp_session_test.datasets["unregistered"][3:]

    for dataset in first_datasets:
        _store_dataset(client, dataset)
        messages.get(timeout=10)

    exp_session = sync_raw.sync_session(*args, **kwargs)
    assert len(exp_session.datasets["registered"]) == 3
    assert len(exp_session.datasets["unregistered"]) == 3

    for dataset in last_datasets:
        _store_dataset(client, dataset)
        messages.get(timeout=10)

    exp_session = sync_raw.sync_session(*args, **kwargs)
    assert len(exp_session.datasets["registered"]) == 6
    assert len(exp_session.datasets["unregistered"]) == 0


def test_unregistered_datasets_content(
    tmp_path: Path, icat_main_client, icat_data_dir: Path
):
    client, messages = icat_main_client
    root_dir = tmp_path

    _ = _init_unregistered_experiment(
        client, messages, icat_data_dir, root_dir, "hg333", "id99"
    )

    exp_session = sync_raw.sync_session(
        client,
        "hg333",
        "id99",
        "20231019",
        root_dir=str(root_dir),
        raw_data_format="esrfv3",
    )

    unregistered = [dset.as_dict() for dset in exp_session.datasets["unregistered"]]

    date_to_string = _astimezone_string(datetime.timezone.utc, iso=True)
    for dataset in unregistered:

        dataset["path"] = os.path.relpath(dataset["path"], str(root_dir))
        dataset["raw_root_dir"] = os.path.relpath(
            dataset["raw_root_dir"], str(root_dir)
        )

        dataset["startdate"] = date_to_string(dataset["startdate"])
        dataset["enddate"] = date_to_string(dataset["enddate"])

        metadata = dataset["metadata"]
        metadata["startDate"] = date_to_string(metadata["startDate"])
        metadata["endDate"] = date_to_string(metadata["endDate"])

    expected = [
        {
            "beamline": "id99",
            "enddate": "2023-10-19T21:00:00+00:00",
            "icat_dataset": None,
            "metadata": {
                "Sample_name": "too_soon_sample",
                "endDate": "2023-10-19T21:00:00+00:00",
                "startDate": "2023-10-19T05:00:00+00:00",
                "investigationId": 0,
            },
            "name": "0001",
            "path": "hg333/id99/20231019/RAW_DATA/too_soon_collection/too_soon_collection_0001",
            "proposal": "hg333",
            "raw_root_dir": "hg333/id99/20231019/RAW_DATA",
            "startdate": "2023-10-19T05:00:00+00:00",
            "status_reason": [
                "started 1:00:00 before the start of the session",
            ],
        },
        {
            "beamline": "id99",
            "enddate": "2023-10-20T16:00:00+00:00",
            "icat_dataset": None,
            "metadata": {
                "Sample_name": "first_sample",
                "endDate": "2023-10-20T16:00:00+00:00",
                "startDate": "2023-10-20T00:00:00+00:00",
                "investigationId": 0,
            },
            "name": "0001",
            "path": "hg333/id99/20231019/RAW_DATA/first_collection/first_collection_0001",
            "proposal": "hg333",
            "raw_root_dir": "hg333/id99/20231019/RAW_DATA",
            "startdate": "2023-10-20T00:00:00+00:00",
            "status_reason": [],
        },
        {
            "beamline": "id99",
            "enddate": "2023-10-21T09:00:00+00:00",
            "icat_dataset": None,
            "metadata": {
                "Sample_name": "first_sample",
                "endDate": "2023-10-21T09:00:00+00:00",
                "startDate": "2023-10-20T17:00:00+00:00",
                "investigationId": 0,
            },
            "name": "0002",
            "path": "hg333/id99/20231019/RAW_DATA/first_collection/first_collection_0002",
            "proposal": "hg333",
            "raw_root_dir": "hg333/id99/20231019/RAW_DATA",
            "startdate": "2023-10-20T17:00:00+00:00",
            "status_reason": [],
        },
        {
            "beamline": "id99",
            "enddate": "2023-10-22T19:00:00+00:00",
            "icat_dataset": None,
            "metadata": {
                "Sample_name": "second_sample",
                "endDate": "2023-10-22T19:00:00+00:00",
                "startDate": "2023-10-22T03:00:00+00:00",
                "investigationId": 0,
            },
            "name": "0001",
            "path": "hg333/id99/20231019/RAW_DATA/second_collection/second_collection_0001",
            "proposal": "hg333",
            "raw_root_dir": "hg333/id99/20231019/RAW_DATA",
            "startdate": "2023-10-22T03:00:00+00:00",
            "status_reason": [],
        },
        {
            "beamline": "id99",
            "enddate": "2023-10-23T12:00:00+00:00",
            "icat_dataset": None,
            "metadata": {
                "Sample_name": "second_sample",
                "endDate": "2023-10-23T12:00:00+00:00",
                "startDate": "2023-10-22T20:00:00+00:00",
                "investigationId": 0,
            },
            "name": "0002",
            "path": "hg333/id99/20231019/RAW_DATA/second_collection/second_collection_0002",
            "proposal": "hg333",
            "raw_root_dir": "hg333/id99/20231019/RAW_DATA",
            "startdate": "2023-10-22T20:00:00+00:00",
            "status_reason": [],
        },
        {
            "beamline": "id99",
            "enddate": "2023-10-24T07:00:00+00:00",
            "icat_dataset": None,
            "metadata": {
                "Sample_name": "too_late_collection",
                "endDate": "2023-10-24T07:00:00+00:00",
                "startDate": "2023-10-23T13:00:00+00:00",
                "investigationId": 0,
            },
            "name": "0001",
            "path": "hg333/id99/20231019/RAW_DATA/too_late_collection/too_late_collection_0001",
            "proposal": "hg333",
            "raw_root_dir": "hg333/id99/20231019/RAW_DATA",
            "startdate": "2023-10-23T13:00:00+00:00",
            "status_reason": [
                "ended 1:00:00 after the end of the session",
            ],
        },
    ]

    for dataset in expected:
        for k in ("path", "raw_root_dir"):
            dataset[k] = dataset[k].replace("/", os.path.sep)

    assert unregistered == expected


def test_missing_files_datasets(
    tmp_path: Path, icat_main_client, icat_add_files_client, icat_data_dir: Path
):
    client, messages = icat_main_client
    client_add, messages_add = icat_add_files_client
    root_dir = tmp_path

    _ = _init_empty_files_experiment(
        client, messages, icat_data_dir, root_dir, "md444", "id99"
    )

    args = client, "md444", "id99", "20230201"
    kwargs = {"root_dir": str(root_dir), "raw_data_format": "esrfv3"}

    exp_session = sync_raw.sync_session(*args, **kwargs)
    assert len(exp_session.datasets["registered"]) == 1
    assert len(exp_session.datasets["registered_without_files"]) == 1
    assert len(exp_session.datasets["unregistered"]) == 0

    for dataset in exp_session.datasets["registered_without_files"]:
        client_add.add_files(dataset.icat_dataset.icat_dataset_id)
        messages_add.get(timeout=10)

    exp_session = sync_raw.sync_session(*args, **kwargs)
    assert len(exp_session.datasets["registered"]) == 2
    assert len(exp_session.datasets["registered_without_files"]) == 0
    assert len(exp_session.datasets["unregistered"]) == 0


def test_excluded_too_many_files(
    tmp_path: Path, icat_main_client, monkeypatch, icat_data_dir: Path
):
    monkeypatch.setattr(sync_raw.sync_types, "MAX_FILES", 10)
    client, messages = icat_main_client
    root_dir = tmp_path

    exp_session = _start_investigation(
        client,
        messages,
        icat_data_dir,
        root_dir,
        "hc555",
        "id99",
        startdate=datetime_utils.deserialize_datetime("2023-02-01T08:00:00+01:00"),
        enddate=datetime_utils.deserialize_datetime("2023-02-06T08:00:00+01:00"),
    )
    dataset_normal = exp_session.create_dataset(
        collection="normal_collection",
        dataset="0001",
        sample_name="normal_sample",
        startdate=datetime_utils.deserialize_datetime("2023-02-01T09:00:00+01:00"),
        duration=datetime.timedelta(hours=2),
        has_content=True,
    )
    _store_dataset(client, dataset_normal)
    messages.get(timeout=10)

    dataset_huge = exp_session.create_dataset(
        collection="huge_collection",
        dataset="0001",
        sample_name="huge_sample",
        startdate=datetime_utils.deserialize_datetime("2023-02-01T12:00:00+01:00"),
        duration=datetime.timedelta(hours=1),
        has_content=True,
    )
    too_many_files = 12
    for i in range(too_many_files):
        with open(os.path.join(dataset_huge.path, f"dummy_{i:04d}.txt"), "w") as f:
            f.write("x" * 10)

    args = client, "hc555", "id99", "20230201"
    kwargs = {"root_dir": str(root_dir), "raw_data_format": "esrfv3"}

    exp_session = sync_raw.sync_session(*args, **kwargs)
    assert len(exp_session.datasets["registered"]) == 1
    assert len(exp_session.datasets["excluded_too_many_files"]) == 1
    assert len(exp_session.datasets["unregistered"]) == 0


def test_session_serialization(tmp_path: Path, icat_main_client, icat_data_dir: Path):
    client, messages = icat_main_client
    root_dir = tmp_path

    _ = _init_unregistered_experiment(
        client, messages, icat_data_dir, root_dir, "hg333", "id99"
    )

    args = client, "hg333", "id99", "20231019"
    kwargs = {"root_dir": str(root_dir), "raw_data_format": "esrfv3"}

    exp_session = sync_raw.sync_session(*args, **kwargs)

    exp_session_copy = exp_session.from_dict(exp_session.as_dict())
    assert exp_session == exp_session_copy


def test_session_pprint(caplog, tmp_path: Path, icat_main_client, icat_data_dir: Path):
    client, messages = icat_main_client
    root_dir = tmp_path

    _ = _init_unregistered_experiment(
        client, messages, icat_data_dir, root_dir, "hg333", "id99"
    )

    args = client, "hg333", "id99", "20231019"
    kwargs = {"root_dir": str(root_dir), "raw_data_format": "esrfv3"}

    exp_session = sync_raw.sync_session(*args, **kwargs)

    for dataset in exp_session.datasets["unregistered"][:3]:
        _store_dataset(client, dataset)
        messages.get(timeout=10)

    exp_session = sync_raw.sync_session(*args, **kwargs)

    date_to_string = _astimezone_string(
        exp_session.icat_investigation.startdate.tzinfo, iso=False
    )
    startdate = date_to_string(exp_session.icat_investigation.startdate)
    enddate = date_to_string(exp_session.icat_investigation.enddate)

    expected = f"""
--------------------------------

Registered datasets:
  /hg333/id99/20231019/RAW_DATA/too_soon_collection/too_soon_collection_0001/
  /hg333/id99/20231019/RAW_DATA/first_collection/first_collection_0001/
  /hg333/id99/20231019/RAW_DATA/first_collection/first_collection_0002/

Invalid datasets:
  /hg333/id99/20231019/RAW_DATA/empty_file/empty_file_0001/
   Cannot extract dataset metadata: HDF5 reading error (HDF5 file not created by Bliss)

Unregistered datasets:
  /hg333/id99/20231019/RAW_DATA/second_collection/second_collection_0001/
  /hg333/id99/20231019/RAW_DATA/second_collection/second_collection_0002/
  /hg333/id99/20231019/RAW_DATA/too_late_collection/too_late_collection_0001/
   ended 1:00:00 after the end of the session

Datasets (TODO):
 3 registered
 1 invalid
 0 not uploaded
 3 unregistered
 0 excluded because of too many files
 0 registered without files

Directory: /hg333/id99/20231019/RAW_DATA/
 Start time: 2023-10-19

Investigation ID: 0
 URL: https://data.esrf.fr/investigation/0/datasets
 Search URL: https://data.esrf.fr/beamline/id99?search=hg333
 Start time: {startdate} <REPLACED>
 End time: {enddate} <REPLACED>
 Duration: 5 days, 0:00:00

--------------------------------
"""

    expected = re.sub(
        r"(/hg333[^\n]*)",
        lambda match: match.group(0).replace("/", os.path.sep),
        expected,
    )

    caplog.set_level(logging.DEBUG, logger="ICAT SYNC")
    exp_session.pprint()
    captured = "\n".join(caplog.messages)
    captured = captured.replace(str(root_dir), "")

    # Remove the sub-strings that depend on the time this test is run
    captured = re.sub(r"\(.*ago\)", "<REPLACED>", captured)

    assert captured == expected


class _ExperimentalSession(sync_types.ExperimentalSession):
    def create_dataset(
        self,
        collection: str,
        dataset: str,
        sample_name: str,
        startdate: datetime.datetime,
        duration: datetime.timedelta,
        status_reason: Tuple[str] = tuple(),
        has_content: bool = True,
    ) -> sync_types.Dataset:
        """Create Bliss dataset directory with HDF5 file. The file contains 3 scans or nothing."""
        with _init_dataset(self.raw_root_dir, collection, dataset) as nxroot:
            dataset_dir = os.path.dirname(nxroot.filename)
            if has_content:
                nxroot.attrs["creator"] = "Bliss"
                nxroot.attrs["file_time"] = startdate.isoformat()

                now, scan_duration, deadtime = _chunk_duration(
                    startdate, startdate + duration, 3
                )

                enddate = _save_scan(nxroot, now, scan_duration, sample_name)

                now += scan_duration + deadtime
                enddate = _save_scan(
                    nxroot, now, scan_duration, sample_name, failed=True
                )

                now += scan_duration + deadtime
                enddate = _save_scan(nxroot, now, scan_duration, sample_name)
            time.sleep(0.050)  # sync_raw sorts by creation time

        if has_content:
            medata_startdate = startdate
            medata_enddate = enddate

            metadata = {
                "Sample_name": sample_name,
                "startDate": medata_startdate,
                "endDate": medata_enddate,
                "investigationId": 0,
            }
        else:
            startdate = None
            enddate = None
            metadata = None

        dataset = sync_types.Dataset(
            path=dataset_dir,
            proposal=self.proposal,
            beamline=self.beamline,
            name=dataset,
            raw_root_dir=self.raw_root_dir,
            status_reason=list(status_reason),
            startdate=startdate,
            enddate=enddate,
            metadata=metadata,
        )
        if has_content:
            assert dataset.get_status() == "unregistered"
        else:
            assert dataset.get_status() == "invalid"
        self.add_dataset(dataset)
        return dataset


def _init_unregistered_experiment(
    client: IcatClient,
    messages,
    icat_data_dir: Path,
    root_dir: Path,
    proposal: str,
    beamline: str,
) -> _ExperimentalSession:
    """Create several datasets on disk:

    - 1 normal dataset starting after the startdate (uregistered)
    - 2 normal datasets (uregistered)
    - 1 dataset with an empty file (uregistered)
    - 2 normal datasets (uregistered)
    - 1 normal dataset ending after the enddate (uregistered)
    """
    startdate = datetime_utils.deserialize_datetime("2023-10-19T08:00:00+02:00")
    enddate = startdate + datetime.timedelta(days=5)

    exp_session = _start_investigation(
        client,
        messages,
        icat_data_dir,
        root_dir,
        proposal,
        beamline,
        startdate,
        enddate,
    )

    now, dataset_duration, deadtime = _chunk_duration(
        startdate, enddate, 7, deadtime=3600
    )

    # Dataset starts before the startdate
    exp_session.create_dataset(
        "too_soon_collection",
        "0001",
        "too_soon_sample",
        now - 2 * deadtime,
        dataset_duration,
        ["started 1:00:00 before the start of the session"],
    )

    now += dataset_duration + deadtime

    # Normal
    _ = exp_session.create_dataset(
        "first_collection", "0001", "first_sample", now, dataset_duration
    )

    now += dataset_duration + deadtime

    # Normal
    _ = exp_session.create_dataset(
        "first_collection", "0002", "first_sample", now, dataset_duration
    )

    now += dataset_duration + deadtime

    # Empty HDF5 file
    _ = exp_session.create_dataset(
        "empty_file",
        "0001",
        "empty_sample",
        now,
        dataset_duration,
        has_content=False,
        status_reason=[
            "Cannot extract dataset metadata: HDF5 reading error (HDF5 file not created by Bliss)"
        ],
    )

    now += dataset_duration + deadtime

    # Normal
    _ = exp_session.create_dataset(
        "second_collection", "0001", "second_sample", now, dataset_duration
    )

    now += dataset_duration + deadtime

    # Normal
    _ = exp_session.create_dataset(
        "second_collection", "0002", "second_sample", now, dataset_duration
    )

    now += dataset_duration + deadtime

    # Dataset ends after the enddate
    _ = exp_session.create_dataset(
        "too_late_collection",
        "0001",
        "too_late_collection",
        now,
        dataset_duration + 2 * deadtime,
        ["ended 1:00:00 after the end of the session"],
    )

    return exp_session


def _init_empty_files_experiment(
    client: IcatClient,
    messages,
    icat_data_dir: Path,
    root_dir: Path,
    proposal: str,
    beamline: str,
):
    """Create several datasets on disk:

    - 1 normal dataset (registered)
    - 1 normal dataset (registered without files)
    """
    startdate = datetime_utils.deserialize_datetime("2023-02-01T08:00:00+01:00")
    enddate = startdate + datetime.timedelta(days=5)

    exp_session = _start_investigation(
        client,
        messages,
        icat_data_dir,
        root_dir,
        proposal,
        beamline,
        startdate,
        enddate,
    )

    now, dataset_duration, deadtime = _chunk_duration(
        startdate, enddate, 2, deadtime=3600
    )

    # Normal
    dataset = exp_session.create_dataset(
        "collection_normal", "0001", "sample_normal", now, dataset_duration
    )
    _store_dataset(client, dataset)
    messages.get(timeout=10)

    now += dataset_duration + deadtime

    # Directory is empty when the file count is calculated (on the ICAT server side)
    dataset = exp_session.create_dataset(
        "collection_missing", "0002", "sample_missing", now, dataset_duration
    )
    with _temporary_empty_directory(dataset.path):
        _store_dataset(client, dataset)
        messages.get(timeout=10)

    return exp_session


@contextmanager
def _temporary_empty_directory(directory_path: str) -> Generator[None, None, None]:
    original_name = path_utils.basename(directory_path)
    new_name = f"{original_name}_tmp"
    renamed_path = os.path.join(path_utils.dirname(directory_path), new_name)
    try:
        os.rename(directory_path, renamed_path)
        os.makedirs(directory_path)
        yield
    finally:
        os.rmdir(directory_path)
        os.rename(renamed_path, directory_path)


def _start_investigation(
    client: IcatClient,
    messages,
    icat_data_dir: Path,
    root_dir: Path,
    proposal: str,
    beamline: str,
    startdate: datetime.datetime,
    enddate: datetime.datetime,
) -> _ExperimentalSession:
    session = startdate.strftime("%Y%m%d")
    session_dir = path_utils.markdir(str(root_dir / proposal / beamline / session))
    raw_root_dir = path_utils.markdir(os.path.join(session_dir, "RAW_DATA"))
    os.makedirs(raw_root_dir, exist_ok=True)

    exp_session = _ExperimentalSession(
        session_dir=session_dir,
        raw_root_dir=raw_root_dir,
        raw_data_format="esrfv3",
        proposal=proposal,
        beamline=beamline,
        session=session,
        startdate=startdate.date(),
        search_url=f"https://data.esrf.fr/beamline/{beamline}?search={proposal}",
        datasets=dict(),
    )

    client.start_investigation(
        beamline=beamline,
        proposal=proposal,
        start_datetime=startdate,
    )
    messages.get(timeout=10)

    _add_investigation_enddate(
        icat_data_dir,
        beamline,
        proposal,
        startdate,
        enddate,
        timeout=10,
    )

    exp_session.icat_investigation = sync_types.IcatInvestigation(
        id=0,
        url="https://data.esrf.fr/investigation/0/datasets",
        search_url=f"https://data.esrf.fr/beamline/{beamline}?search={proposal}",
        registered_datasets=list(),
        startdate=startdate,
        enddate=enddate,
    )

    return exp_session


def _add_investigation_enddate(
    icat_data_dir: Path,
    beamline: str,
    proposal: str,
    startdate: datetime.datetime,
    enddate: datetime.datetime,
    timeout: float = 10.0,
    interval: float = 1.0,
) -> None:
    """Add endDate to an investigation entry created with start_investigation.

    Retries until the investigation appears or timeout is reached.
    """
    investigations_file = icat_data_dir / "investigations.json"
    deadline = time.monotonic() + timeout

    while True:
        try:
            with investigations_file.open() as f:
                investigations = json.load(f)
        except (FileNotFoundError, json.JSONDecodeError):
            investigations = None

        if investigations:
            inv = investigations[-1]
            if (
                inv.get("proposal") == proposal
                and inv.get("instrument", {}).get("name") == beamline
            ):
                inv["endDate"] = enddate.isoformat()

                with investigations_file.open("w") as f:
                    json.dump(investigations, f)

                return

        if time.monotonic() >= deadline:
            pprint(investigations)
            raise TimeoutError(
                f"Investigation not found within {timeout}s for "
                f"proposal={proposal}, beamline={beamline}, "
                f"startDate={startdate.isoformat()}"
            )

        time.sleep(interval)


def _store_dataset(client: IcatClient, dataset: sync_types.Dataset) -> None:
    client.store_dataset(
        beamline=dataset.beamline,
        proposal=dataset.proposal,
        dataset=dataset.name,
        path=dataset.path,
        metadata=dataset.metadata,
    )


def _chunk_duration(
    startdate: datetime.datetime,
    enddate: datetime.datetime,
    nchunks: int,
    deadtime: int = 0,
):
    deadtime = datetime.timedelta(seconds=deadtime)
    total_deadtime = (nchunks + 1) * deadtime
    total_time = enddate - startdate
    chunk_duration = (total_time - total_deadtime) / nchunks
    now = startdate + deadtime
    return now, chunk_duration, deadtime


def _save_scan(
    nxroot: h5py.File,
    startdate: datetime.datetime,
    duration: datetime.timedelta,
    sample_name: str,
    failed: bool = False,
) -> Optional[datetime.datetime]:
    scans = list(nxroot)
    if scans:
        scan = max(map(int, map(float, scans))) + 1
    else:
        scan = 1
    name = f"{scan}.1"
    grp = nxroot.create_group(name)

    grp["start_time"] = startdate.isoformat()
    if failed:
        return

    enddate = startdate + duration
    grp["end_time"] = enddate.isoformat()
    grp["sample/name"] = sample_name
    return enddate


@contextmanager
def _init_dataset(
    raw_root_dir: str,
    collection: str,
    dataset: str,
) -> Generator[h5py.File, None, None]:
    basename = f"{collection}_{dataset}"
    dataset_dir = os.path.join(raw_root_dir, collection, basename)
    os.makedirs(dataset_dir, exist_ok=True)

    filename = os.path.basename(dataset_dir) + ".h5"
    with h5py.File(os.path.join(dataset_dir, filename), mode="w") as f:
        yield f


def _astimezone_string(tzinfo, iso: bool = True) -> Callable[[datetime.datetime], str]:
    def astimezone_string(dt: datetime.datetime) -> str:
        dt = dt.astimezone(tz=tzinfo)
        if iso:
            return dt.isoformat()
        return str(dt)

    return astimezone_string



import datetime
from numbers import Number
from typing import Any
from typing import Dict
from typing import Sequence

from .... import errors
from ...utils import datetime_utils


def fallback_serialization(
    parameters: Dict[str, Any],
) -> Dict[str, str]:
    """Naive parameter value serialization used when Pydantic validation fails and `strict=False`.

    :raises IcatMetadataSerializationError:
    """
    return {
        name: _serialize_parameter(name, value)
        for name, value in parameters.items()
        if value is not None
    }


def _serialize_parameter(name: str, value: Any, _depth: int = 0) -> str:
    if isinstance(value, str):
        return value
    elif isinstance(value, bytes):
        return value.decode()
    elif isinstance(value, bool):
        return "true" if value else "false"
    elif isinstance(value, Number):
        return str(value)
    elif isinstance(value, datetime.datetime):
        return datetime_utils.with_timezone_zulu_format(value)
    elif isinstance(value, Sequence):
        if _depth > 1:
            raise errors.IcatMetadataSerializationError(
                f"Invalid ICAT dataset parameter ({name}={value}): sequences can only be nested 2-levels deep"
            )

        if len({type(v) for v in value if v is not None}) > 1:
            raise errors.IcatMetadataSerializationError(
                f"Invalid ICAT dataset parameter ({name}={value}): sequences cannot have mixed types"
            )

        if all(isinstance(v, (str, bytes, Number)) or v is None for v in value):
            sep = ","
        else:
            sep = " "

        return sep.join(
            [
                _serialize_parameter(f"{name}[{i}]", v, _depth + 1)
                for i, v in enumerate(value)
                if v is not None
            ]
        )
    else:
        raise errors.IcatMetadataSerializationError(
            f"Invalid ICAT dataset parameter ({name}={value}): unsupported type {type(value)}"
        )



from dataclasses import dataclass
from dataclasses import field
from typing import Optional

from xsdata.models.datatype import XmlDateTime

__NAMESPACE__ = "http://www.esrf.fr/icat"


@dataclass
class Investigation:
    class Meta:
        name = "investigation"
        namespace = "http://www.esrf.fr/icat"

    experiment: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    instrument: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
        },
    )
    start_date: Optional[XmlDateTime] = field(
        default=None,
        metadata={
            "name": "startDate",
            "type": "Element",
            "required": True,
        },
    )
    investigation_id: Optional[int] = field(
        default=None,
        metadata={
            "name": "investigationId",
            "type": "Element",
            "required": True,
        },
    )



import logging
import pathlib
import sys

logger = logging.getLogger(__name__)

if __name__ == "__main__":
    try:
        import pkg_resources  # noqa F401
    except ImportError:
        sys.path.insert(0, str(pathlib.Path(__file__).parent))

    from coilmq.start import main

    try:
        main()
    except (KeyboardInterrupt, SystemExit):
        pass
    except Exception as e:
        logger.error("Server terminated due to error: %s" % e)
        logger.exception(e)



import datetime
import warnings
from typing import List
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import numpy

from . import defaults
from .add_files import IcatAddFilesClient
from .deprecation_utils import deprecated_argument
from .deprecation_utils import deprecated_property
from .elogbook import IcatElogbookClient
from .icatplus_restricted import IcatPlusRestrictedClient
from .investigation import IcatInvestigationClient
from .metadata import IcatMetadataClient
from .types import ArchiveStatusLevel
from .types import ArchiveStatusLevel as StatusLevel  # noqa F401 deprecated
from .types import ArchiveStatusType
from .types import ArchiveStatusType as StatusType  # noqa F401 deprecated
from .types import Dataset
from .types import DatasetId
from .types import SampleMetadata
from .update_metadata import IcatUpdateMetadataClient

with warnings.catch_warnings():
    warnings.simplefilter("ignore", category=ImportWarning)
    from .archive import IcatArchiveStatusClient
    from .interface import IcatClientInterface


class IcatClient(IcatClientInterface):
    """Client object that provides access to these services:

    - ActiveMQ message broker for creating datasets in ICAT
    - ActiveMQ message broker for updating dataset metadata in ICAT
    - ActiveMQ message broker for updating dataset file count in ICAT
    - ActiveMQ message broker for updating dataset archiving status in ICAT (DEPRECATED)
    - RESTful interface for sending electronic logbook messages/images and get information about investigations

    The RESTful interface is referred to as ICAT+ and the ActiveMQ message brokers are consumed by the "ingesters".
    """

    def __init__(
        self,
        metadata_urls: Optional[List[str]] = None,
        elogbook_url: Optional[str] = None,
        elogbook_token: Optional[str] = None,
        metadata_queue: Optional[str] = None,
        metadata_queue_monitor_port: Optional[int] = None,
        elogbook_timeout: Optional[float] = None,
        feedback_timeout: Optional[float] = None,
        queue_timeout: Optional[float] = None,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        elogbook_metadata: Optional[Mapping] = None,
        archive_urls: Optional[List[str]] = None,  # DEPRECATED
        archive_queue: Optional[str] = None,  # DEPRECATED
        archive_queue_monitor_port: Optional[int] = None,  # DEPRECATED
        update_metadata_urls: Optional[List[str]] = None,  # DEPRECATED
        update_metadata_queue: Optional[str] = None,
        update_metadata_queue_monitor_port: Optional[int] = None,
        add_files_urls: Optional[List[str]] = None,
        add_files_queue: Optional[str] = None,
        add_files_queue_monitor_port: Optional[int] = None,
        reschedule_investigation_urls: Optional[List[str]] = None,
        reschedule_investigation_queue: Optional[str] = None,
        reschedule_investigation_queue_monitor_port: Optional[int] = None,
        icatplus_restricted_url: Optional[str] = None,
        icatplus_password: Optional[str] = None,
        icat_session_id: Optional[str] = None,
        catalogue_queues: Optional[List[str]] = None,  # DEPRECATED
        catalogue_url: Optional[str] = None,  # DEPRECATED
        tracking_url: Optional[str] = None,  # DEPRECATED
    ):
        """
        :param metadata_urls: URLs of the ActiveMQ message brokers to be used for creating ICAT datasets from a directory with metadata.
        :param elogbook_url: URL of the ICAT+ REST server to be used for sending text or images to the electronic logbook and get information about investigations.
        :param elogbook_token: Access token for restricted requests to `elogbook_url`.
        :param metadata_queue: Queue to be used when sending a message to `metadata_urls`.
        :param metadata_queue_monitor_port: REST server port to be used for monitor the `metadata_urls` ActiveMQ message brokers (same host as the message broker).
        :param elogbook_timeout: POST timeout for `elogbook_url`.
        :param feedback_timeout: GET timeout for `elogbook_url`.
        :param queue_timeout: Connection timeout for the ActiveMQ message brokers.
        :param beamline: Default beamline to be used as metadata when sending messages to `metadata_urls` or `elogbook_url`.
        :param proposal: Default proposal to be used as metadata when sending messages to `metadata_urls` or `elogbook_url`.
        :param elogbook_metadata: Default electronic logbook metadata to be used when sending messages to  `elogbook_url`.
        :param archive_urls: URLs of the ActiveMQ message brokers to be used for updating the archival status of ICAT datasets (DEPRECATED - This argument is deprecated and will be removed in a future version).
        :param archive_queue: Queue to be used when sending a message to `archive_urls` (DEPRECATED - This argument is deprecated and will be removed in a future version).
        :param archive_queue_monitor_port: REST server port to be used for monitor the `archive_urls` ActiveMQ message brokers (same host as the message broker) (DEPRECATED - This argument is deprecated and will be removed in a future version).
        :param update_metadata_urls: URLs of the ActiveMQ message brokers to be used for update metadata of ICAT datasets.
        :param update_metadata_queue: Queue to be used when sending a message to `update_metadata_urls`.
        :param update_metadata_queue_monitor_port: REST server port to be used for monitor the `update_metadata_urls` ActiveMQ message brokers (same host as the message broker).
        :param add_files_urls: URLs of the ActiveMQ message brokers to be used for updating the file count of ICAT datasets.
        :param add_files_queue: Queue to be used when sending a message to `add_files_urls`.
        :param add_files_queue_monitor_port: REST server port to be used for monitor the `add_files_urls` ActiveMQ message brokers (same host as the message broker).
        :param reschedule_investigation_urls: URLs of the ActiveMQ message brokers to be used for rescheduling investigations.
        :param reschedule_investigation_queue: Queue to be used when sending a message to `reschedule_investigation`.
        :param reschedule_investigation_queue_monitor_port: REST server port to be used for monitor the `reschedule_investigation` ActiveMQ message brokers (same host as the message broker).
        :param icatplus_restricted_url: URL of the ICAT+ REST server to be used for restricted access (requires `icatplus_password` or `do_log_in`).
        :param icatplus_password: Password to provide access to `icatplus_restricted_url`.
        :param icat_session_id: The ICAT's session id that is used for authentication/authorization in ICAT.
        :param catalogue_queues: URLs of the ActiveMQ message brokers to be used for the catalogue (DEPRECATED).
        :param catalogue_url: URL of the ICAT+ REST server to be used for accessing the catalogue (DEPRECATED).
        :param tracking_url: URL of the ICAT+ REST server to be used for accessing the tracking (DEPRECATED).
        """

        # Defaults to be used in client methods
        self.current_proposal = proposal
        self.current_beamline = beamline
        self.current_dataset = None
        self.current_path = None
        self.current_dataset_metadata = None

        # Deprecated constructor arguments
        if catalogue_queues:
            reschedule_investigation_urls = catalogue_queues
            warnings.warn(
                "'catalogue_queues' is deprecated, use 'reschedule_investigation_urls' instead.",
                DeprecationWarning,
                stacklevel=2,
            )
        if catalogue_url:
            warnings.warn(
                "'catalogue_url' is deprecated, use 'icatplus_restricted_url instead'.",
                DeprecationWarning,
                stacklevel=2,
            )
            if not icatplus_restricted_url:
                icatplus_restricted_url = catalogue_url
        if tracking_url:
            warnings.warn(
                "'tracking_url' is deprecated, use 'icatplus_restricted_url' instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            if not icatplus_restricted_url:
                icatplus_restricted_url = tracking_url

        # Initialize clients to message brokers or REST URL's.
        # Clients with missing parameters are not initialized
        # add the associated client methods will raise an error.
        self._init_metadata_client(
            brokers=metadata_urls,
            queue=metadata_queue,
            timeout=queue_timeout,
            monitor_port=metadata_queue_monitor_port,
        )
        self._init_archive_client(
            brokers=archive_urls,
            queue=archive_queue,
            timeout=queue_timeout,
            monitor_port=archive_queue_monitor_port,
        )
        self._init_update_metadata_client(
            brokers=update_metadata_urls,
            queue=update_metadata_queue,
            timeout=queue_timeout,
            monitor_port=update_metadata_queue_monitor_port,
        )
        self._init_elogbook_client(
            url=elogbook_url,
            api_key=elogbook_token,
            timeout=elogbook_timeout,
            metadata=elogbook_metadata,
        )
        self._init_investigation_client(
            url=elogbook_url,
            api_key=elogbook_token,
            timeout=feedback_timeout,
        )
        self._init_add_files_client(
            brokers=add_files_urls,
            queue=add_files_queue,
            timeout=queue_timeout,
            monitor_port=add_files_queue_monitor_port,
        )
        self._init_reschedule_investigation_client(
            brokers=reschedule_investigation_urls,
            queue=reschedule_investigation_queue,
            timeout=queue_timeout,
            monitor_port=reschedule_investigation_queue_monitor_port,
        )
        self._init_icatplus_restricted_client(
            url=icatplus_restricted_url,
            password=icatplus_password,
            session_id=icat_session_id,
        )

    def disconnect(self):
        if self.__metadata_client is not None:
            self.__metadata_client.disconnect()
        if self.__update_metadata_client is not None:
            self.__update_metadata_client.disconnect()
        if self.__add_files_client is not None:
            self.__add_files_client.disconnect()
        if self.__archive_client is not None:
            self.__archive_client.disconnect()
        if self.__reschedule_investigation_client is not None:
            self.__reschedule_investigation_client.disconnect()

    @deprecated_property
    def metadata_client(self):
        return self._metadata_client

    @property
    def _metadata_client(self):
        if self.__metadata_client is None:
            raise RuntimeError("The message queue URL's are missing")
        return self.__metadata_client

    def _init_metadata_client(
        self,
        brokers: Optional[List[str]] = None,
        queue: Optional[str] = None,
        timeout: Optional[float] = None,
        monitor_port: Optional[int] = None,
    ):
        if brokers:
            self.__metadata_client = IcatMetadataClient(
                queue_urls=brokers,
                queue_name=queue,
                monitor_port=monitor_port,
                timeout=timeout,
            )
        else:
            self.__metadata_client = None

    @deprecated_property
    def elogbook_client(self):
        return self._elogbook_client

    @property
    def _elogbook_client(self):
        if self.__elogbook_client is None:
            raise RuntimeError("The ICAT+ URL and/or token are missing")
        return self.__elogbook_client

    def _init_elogbook_client(
        self,
        url: Optional[str] = None,
        api_key: Optional[str] = None,
        timeout: Optional[float] = None,
        metadata: Optional[dict] = None,
    ):
        if url and api_key:
            if metadata is None:
                metadata = dict()
            self.__elogbook_client = IcatElogbookClient(
                url=url,
                api_key=api_key,
                timeout=timeout,
                **metadata,
            )
        else:
            self.__elogbook_client = None

    @deprecated_property
    def investigation_client(self):
        return self._investigation_client

    @property
    def _investigation_client(self):
        if self.__investigation_client is None:
            raise RuntimeError("The ICAT+ URL and/or token are missing")
        return self.__investigation_client

    def _init_investigation_client(
        self,
        url: Optional[str] = None,
        api_key: Optional[str] = None,
        timeout: Optional[float] = None,
    ):
        if url and api_key:
            self.__investigation_client = IcatInvestigationClient(
                url=url, api_key=api_key, timeout=timeout
            )
        else:
            self.__investigation_client = None

    @deprecated_property
    def archive_client(self):
        return self._archive_client

    @property
    def _archive_client(self):
        if self.__archive_client is None:
            raise RuntimeError("The message queue URL's are missing")
        return self.__archive_client

    def _init_archive_client(
        self,
        brokers: Optional[List[str]] = None,
        queue: Optional[str] = None,
        timeout: Optional[float] = None,
        monitor_port: Optional[int] = None,
    ):
        if brokers:
            warnings.warn(
                "'brokers' (from 'archive_urls') is deprecated and will be removed in a future version.",
                DeprecationWarning,
                stacklevel=2,
            )
            self.__archive_client = IcatArchiveStatusClient(
                queue_urls=brokers,
                queue_name=queue,
                monitor_port=monitor_port,
                timeout=timeout,
            )
        else:
            self.__archive_client = None

    @deprecated_property
    def update_metadata_client(self):
        return self._update_metadata_client

    @property
    def _update_metadata_client(self):
        if self.__update_metadata_client is None:
            raise RuntimeError("The message queue URL's are missing")
        return self.__update_metadata_client

    def _init_update_metadata_client(
        self,
        brokers: Optional[List[str]] = None,
        queue: Optional[str] = None,
        timeout: Optional[float] = None,
        monitor_port: Optional[int] = None,
    ):
        if brokers:
            self.__update_metadata_client = IcatUpdateMetadataClient(
                queue_urls=brokers,
                queue_name=queue,
                monitor_port=monitor_port,
                timeout=timeout,
            )
        else:
            self.__update_metadata_client = None

    @deprecated_property
    def add_files_client(self):
        return self._add_files_client

    @property
    def _add_files_client(self):
        if self.__add_files_client is None:
            raise RuntimeError("The message queue URL's are missing")
        return self.__add_files_client

    def _init_add_files_client(
        self,
        brokers: Optional[List[str]] = None,
        queue: Optional[str] = None,
        timeout: Optional[float] = None,
        monitor_port: Optional[int] = None,
    ):
        if brokers:
            self.__add_files_client = IcatAddFilesClient(
                queue_urls=brokers,
                queue_name=queue,
                monitor_port=monitor_port,
                timeout=timeout,
            )
        else:
            self.__add_files_client = None

    @deprecated_property
    def catalogue_client(self):
        return self

    @deprecated_property
    def tracking_client(self):
        return self

    @property
    def _icatplus_authentication_client(self):
        if self.__icatplus_authentication_client is None:
            raise RuntimeError("The ICAT+ URL and/or token are missing")
        return self.__icatplus_authentication_client

    @property
    def _icatplus_restricted_client(self):
        if self.__icatplus_restricted_client is None:
            raise RuntimeError("The ICAT+ URL is missing and/or login")
        return self.__icatplus_restricted_client

    def _init_icatplus_restricted_client(
        self,
        url: Optional[str] = None,
        password: Optional[str] = None,
        session_id: Optional[str] = None,
    ):
        if url:
            self.__icatplus_restricted_client = IcatPlusRestrictedClient(
                url=url, password=password, session_id=session_id
            )
        else:
            self.__icatplus_restricted_client = None

    @property
    def _reschedule_investigation_client(self):
        if self.__reschedule_investigation_client is None:
            raise RuntimeError("The message queue URL's are missing")
        return self.__reschedule_investigation_client

    def _init_reschedule_investigation_client(
        self,
        brokers: Optional[List[str]] = None,
        queue: Optional[str] = None,
        timeout: Optional[float] = None,
        monitor_port: Optional[int] = None,
    ):
        if brokers:
            self.__reschedule_investigation_client = IcatMetadataClient(
                queue_urls=brokers or defaults.RESCHEDULE_INVESTIGATION_BROKERS,
                queue_name=queue or defaults.RESCHEDULE_INVESTIGATION_QUEUE,
                monitor_port=monitor_port,
                timeout=timeout,
            )
        else:
            self.__reschedule_investigation_client = None

    @property
    def current_proposal(self):
        return self.__current_proposal

    @current_proposal.setter
    def current_proposal(self, value: Optional[str]):
        self.__current_proposal = value

    @property
    def current_beamline(self):
        return self.__current_beamline

    @current_beamline.setter
    def current_beamline(self, value: Optional[str]):
        self.__current_beamline = value

    @property
    def current_dataset(self):
        return self.__current_dataset

    @current_dataset.setter
    def current_dataset(self, value: Optional[str]):
        self.__current_dataset = value

    @property
    def current_dataset_metadata(self):
        return self.__current_dataset_metadata

    @current_dataset_metadata.setter
    def current_dataset_metadata(self, value: Optional[dict]):
        self.__current_dataset_metadata = value

    @property
    def current_path(self):
        return self.__current_path

    @current_path.setter
    def current_path(self, value: Optional[str]):
        self.__current_path = value

    def send_message(
        self,
        msg: str,
        msg_type: Optional[str] = None,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        investigation_id: Optional[str] = None,
        dataset: Optional[str] = None,
        beamline_only: Optional[bool] = None,
        editable: Optional[bool] = None,
        formatted: Optional[bool] = None,
        mimetype: Optional[str] = None,
        **payload,
    ):
        if beamline_only:
            proposal = None
        elif proposal is None:
            proposal = self.current_proposal
        if beamline is None:
            beamline = self.current_beamline
        if beamline_only:
            dataset = None
        elif dataset is None:
            dataset = self.current_dataset
        self._elogbook_client.send_message(
            message=msg,
            message_type=msg_type,
            beamline=beamline,
            proposal=proposal,
            investigation_id=investigation_id,
            dataset=dataset,
            editable=editable,
            formatted=formatted,
            mimetype=mimetype,
            **payload,
        )

    def send_binary_data(
        self,
        data: bytes,
        mimetype: Optional[str] = None,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        investigation_id: Optional[str] = None,
        beamline_only: Optional[bool] = None,
        **payload,
    ):
        if beamline_only:
            proposal = None
        elif proposal is None:
            proposal = self.current_proposal
        if beamline is None:
            beamline = self.current_beamline
        self._elogbook_client.send_binary_data(
            data,
            mimetype=mimetype,
            beamline=beamline,
            proposal=proposal,
            investigation_id=investigation_id,
            **payload,
        )

    def send_text_file(
        self,
        filename: str,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        dataset: Optional[str] = None,
        beamline_only: Optional[bool] = None,
        **payload,
    ):
        if beamline_only:
            proposal = None
        elif proposal is None:
            proposal = self.current_proposal
        if beamline is None:
            beamline = self.current_beamline
        if beamline_only:
            dataset = None
        elif dataset is None:
            dataset = self.current_dataset
        self._elogbook_client.send_text_file(
            filename, beamline=beamline, proposal=proposal, dataset=dataset, **payload
        )

    def send_binary_file(
        self,
        filename: str,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        beamline_only: Optional[bool] = None,
        **payload,
    ):
        if beamline_only:
            proposal = None
        elif proposal is None:
            proposal = self.current_proposal
        if beamline is None:
            beamline = self.current_beamline
        self._elogbook_client.send_binary_file(
            filename, beamline=beamline, proposal=proposal, **payload
        )

    def start_investigation(
        self,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        start_datetime=None,
        end_datetime=None,
    ):
        if proposal is None:
            proposal = self.current_proposal
        else:
            self.current_proposal = proposal
        if beamline is None:
            beamline = self.current_beamline
        else:
            self.current_beamline = beamline
        self._metadata_client.start_investigation(
            beamline=beamline,
            proposal=proposal,
            start_datetime=start_datetime,
            end_datetime=end_datetime,
        )

    def store_dataset(
        self,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        dataset: Optional[str] = None,
        path: Optional[str] = None,
        metadata: dict = None,
        store_filename: Optional[str] = None,
        strict: bool = False,
    ):
        """
        Send raw dataset (i.e. does not depend on other datasets) metadata to ICAT or file.

        :param beamline: The beamline name of the proposal.
        :param proposal: The proposal name.
        :param dataset: The dataset name.
        :param path: The path to the dataset on disk.
        :param metadata: A dictionary of metadata to be attached to the dataset.
        :param store_filename: Save metadata in this file instead of sending it to ICAT.
        :param strict: Strict metadata validation.

        :raises IcatMetadataSerializationError: If metadata serialization fails.
        :raises IcatMetadataValidationError: If ``strict=True`` and the metadata is invalid.
        :emits IcatMetadataValidationWarning: If ``strict=False`` and the metadata is invalid.
        :emits IcatMetadataDeprecatedWarning: If metadata is deprecated.
        """
        if proposal is None:
            proposal = self.current_proposal
        if beamline is None:
            beamline = self.current_beamline
        if dataset is None:
            dataset = self.current_dataset
        if path is None:
            path = self.current_path
        if metadata is None:
            metadata = self.current_dataset_metadata
            if metadata is None:
                metadata = dict()
        if store_filename:
            self._metadata_client.store_metadata(
                store_filename,
                beamline=beamline,
                proposal=proposal,
                dataset=dataset,
                path=path,
                metadata=metadata,
                strict=strict,
            )
        else:
            self._metadata_client.send_metadata(
                beamline=beamline,
                proposal=proposal,
                dataset=dataset,
                path=path,
                metadata=metadata,
                strict=strict,
            )

    def store_processed_data(
        self,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        dataset: Optional[str] = None,
        path: Optional[str] = None,
        metadata: dict = None,
        raw: Sequence = tuple(),
        store_filename: Optional[str] = None,
        strict: bool = False,
    ):
        """
        Send processed dataset (i.e. depends on other datasets) metadata to ICAT or file.

        :param beamline: The beamline name of the proposal.
        :param proposal: The proposal name.
        :param dataset: The dataset name.
        :param path: The path to the dataset on disk.
        :param metadata: A dictionary of metadata to be attached to the dataset.
        :param raw: shorthand for ``metadata = {'input_datasets': ...}``.
        :param store_filename: Save metadata in this file instead of sending it to ICAT.
        :param strict: Strict metadata validation.

        :raises IcatMetadataSerializationError: If metadata serialization fails.
        :raises IcatMetadataValidationError: If ``strict=True`` and the metadata is invalid.
        :emits IcatMetadataValidationWarning: If ``strict=False`` and the metadata is invalid.
        :emits IcatMetadataDeprecatedWarning: If metadata is deprecated.
        """
        if metadata is None:
            metadata = self.current_dataset_metadata
            if metadata is None:
                metadata = dict()
        if raw:
            if isinstance(raw, str):
                metadata["input_datasets"] = [raw]
            elif isinstance(raw, Sequence):
                metadata["input_datasets"] = list(raw)
            else:
                metadata["input_datasets"] = [raw]
        if not metadata.get("input_datasets"):
            raise ValueError("Provide 'raw' dataset directories")
        self.store_dataset(
            beamline=beamline,
            proposal=proposal,
            dataset=dataset,
            path=path,
            metadata=metadata,
            store_filename=store_filename,
            strict=strict,
        )

    def store_dataset_from_file(self, store_filename: Optional[str] = None):
        self._metadata_client.send_metadata_from_file(store_filename)

    def investigation_info(
        self,
        beamline: str,
        proposal: str,
        date: Optional[Union[datetime.datetime, datetime.date]] = None,
        allow_open_ended: bool = True,
        timeout: Optional[float] = None,
    ) -> Optional[dict]:
        return self._investigation_client.investigation_info(
            beamline=beamline,
            proposal=proposal,
            date=date,
            allow_open_ended=allow_open_ended,
            timeout=timeout,
        )

    def registered_dataset_ids(
        self,
        beamline: str,
        proposal: str,
        date: Optional[Union[datetime.datetime, datetime.date]] = None,
        allow_open_ended: bool = True,
        timeout: Optional[float] = None,
    ) -> Optional[List[DatasetId]]:
        return self._investigation_client.registered_dataset_ids(
            beamline=beamline,
            proposal=proposal,
            date=date,
            allow_open_ended=allow_open_ended,
            timeout=timeout,
        )

    def registered_datasets(
        self,
        beamline: str,
        proposal: str,
        date: Optional[Union[datetime.datetime, datetime.date]] = None,
        allow_open_ended: bool = True,
        timeout: Optional[float] = None,
    ) -> Optional[List[Dataset]]:
        return self._investigation_client.registered_datasets(
            beamline=beamline,
            proposal=proposal,
            date=date,
            allow_open_ended=allow_open_ended,
            timeout=timeout,
        )

    def investigation_info_string(
        self,
        beamline: str,
        proposal: str,
        date: Optional[Union[datetime.datetime, datetime.date]] = None,
        allow_open_ended: bool = True,
        timeout: Optional[float] = None,
    ) -> str:
        info = self.investigation_info(
            beamline=beamline,
            proposal=proposal,
            date=date,
            allow_open_ended=allow_open_ended,
            timeout=timeout,
        )
        if info:
            rows = [(str(k), str(v)) for k, v in info.items()]
            lengths = numpy.array([[len(s) for s in row] for row in rows])
            fmt = "   ".join(["{{:<{}}}".format(n) for n in lengths.max(axis=0)])
            infostr = "ICAT proposal time slot:\n "
            infostr += "\n ".join([fmt.format(*row) for row in rows])
        elif info is None:
            infostr = f"Proposal information currently not available ({self.reason_for_missing_information})"
        else:
            infostr = "Proposal NOT available in the data portal"
        return infostr

    def investigation_summary(
        self,
        beamline: str,
        proposal: str,
        date: Optional[Union[datetime.datetime, datetime.date]] = None,
        allow_open_ended: bool = True,
        timeout: Optional[float] = None,
    ) -> List[Tuple]:
        info = self.investigation_info(
            beamline=beamline,
            proposal=proposal,
            date=date,
            allow_open_ended=allow_open_ended,
            timeout=timeout,
        )
        keys = ["e-logbook", "data portal"]
        if info:
            rows = [(key, info[key]) for key in keys]
        elif info is None:
            rows = [
                (
                    key,
                    f"Proposal information currently not available ({self.reason_for_missing_information})",
                )
                for key in keys
            ]
        else:
            rows = [(key, "Proposal NOT available in the data portal") for key in keys]
        return rows

    def update_archive_restore_status(
        self,
        dataset_id: int = None,
        type: ArchiveStatusType = None,
        level: ArchiveStatusLevel = ArchiveStatusLevel.INFO,
        message: Optional[str] = None,
    ):
        self._archive_client.send_archive_status(
            dataset_id=dataset_id, type=type, level=level, message=message
        )

    def update_metadata(
        self,
        proposal: str = None,
        beamline: str = None,
        dataset_paths: str = None,
        metadata_name: str = None,
        metadata_value: str = None,
    ):
        if proposal is None:
            proposal = self.current_proposal
        if beamline is None:
            beamline = self.current_beamline
        self.__update_metadata_client.send_update_metadata(
            proposal=proposal,
            beamline=beamline,
            dataset_paths=dataset_paths,
            metadata_name=metadata_name,
            metadata_value=metadata_value,
        )

    def add_files(
        self,
        dataset_id: int = None,
    ):
        self.__add_files_client.add_files(
            dataset_id=dataset_id,
        )

    def reschedule_investigation(self, investigation_id: str):
        self._reschedule_investigation_client.reschedule_investigation(investigation_id)

    def do_log_in(
        self, password: str, username: Optional[str] = None, plugin: str = "esrf"
    ) -> dict:
        return self._icatplus_restricted_client.login(password, username, plugin)

    def get_investigations_by(
        self,
        filter: Optional[str] = None,
        instrument_name: Optional[str] = None,
        start_date: Optional[datetime.datetime] = None,
        end_date: Optional[datetime.datetime] = None,
        ids: Optional[str] = None,
    ) -> List[dict]:
        return self._icatplus_restricted_client.get_investigations_by(
            filter=filter,
            instrument_name=instrument_name,
            start_date=start_date,
            end_date=end_date,
            ids=ids,
        )

    def get_datasets_by(
        self,
        investigation_id: Optional[str] = None,
        dataset_ids: Optional[str] = None,
    ) -> List[dict]:

        return self._icatplus_restricted_client.get_datasets_by(
            investigation_id=investigation_id, dataset_ids=dataset_ids
        )

    def get_samples_by(
        self,
        investigation_id: Optional[str] = None,
        sample_ids: Optional[str] = None,
        investigationId: Optional[str] = None,  # noqa N803
        sampleIds: Optional[str] = None,  # noqa N803
    ) -> List[dict]:
        investigation_id = deprecated_argument(
            "investigation_id", investigation_id, "investigationId", investigationId
        )
        sample_ids = deprecated_argument(
            "sample_ids", sample_ids, "sampleIds", sampleIds
        )
        return self._icatplus_restricted_client.get_samples_by(
            investigation_id=investigation_id, sample_ids=sample_ids
        )

    def get_parcels_by(self, investigation_id: str) -> List[dict]:
        return self._icatplus_restricted_client.get_parcels_by(investigation_id)

    def get_session_information(self) -> dict:
        return self._icatplus_restricted_client.get_session_information()

    def get_sample_metadata_by(
        self,
        proposal: str,
        beamline: str,
        session_start_date: Optional[datetime.date] = None,
    ) -> List[SampleMetadata]:
        return self._icatplus_restricted_client.get_sample_metadata_by(
            proposal, beamline, session_start_date
        )

    def get_sample_files_information_by(self, sample_id: str) -> dict:
        return self._icatplus_restricted_client.get_sample_files_information_by(
            sample_id
        )

    def download_file_by(
        self,
        sample_id: str,
        resource_id: str,
        use_chunks: bool = False,
        chunk_size: int = 8192,
    ) -> bytes:
        return self._icatplus_restricted_client.download_file_by(
            sample_id, resource_id, use_chunks, chunk_size
        )

    @deprecated_property
    def expire_datasets_on_close(self) -> bool:
        return False

    @deprecated_property
    def reason_for_missing_information(self) -> str:
        return "ICAT communication timeout"
